'''Test the higher level behavior of the unit commitment'''
import random
from minpower.generators import Generator
import pandas as pd
from pandas.util.testing import assert_frame_equal, assert_series_equal
from test_utils import *

@istest
@with_setup(setup=get_duals, teardown=reset_config)
def prices():
    '''
    Run a basic unit commitment. 
    Ensure that the correct LMPs are returned for all times.
    '''
    generators=[
        make_cheap_gen(pmax=100),
        make_mid_gen(pmax=20),
        make_expensive_gen()
    ]

    power_system, times = solve_problem(generators, do_reset_config=False,
        **make_loads_times(Pdt=[80,110,130]))
    lmps = [power_system.buses[0].price(t) for t in times]
    
    assert lmps==[gen_costs['cheap'],gen_costs['mid'],gen_costs['expensive']]
        
@istest
def rolling():
    '''
    Run a basic unit commitment over 72hrs
    Ensure that the generation meets the load for each time.
    '''
    generators=[Generator(costcurveequation='10P+.01P^2')]
    Pdt=[random.randrange(0, 200) for i in range(0,72)]
    power_system,times=solve_problem(generators,**make_loads_times(Pdt=Pdt))
    load=power_system.loads()[0]
    load_balanced = all(generators[0].power(t)==load.power(t) for t in times)
    assert load_balanced


@istest
@with_setup(teardown=reset_config)
def load_shedding():
    '''
    Create a single generator with a high limit.
    Create a load that exceeds that limit at t1.
    Ensure that:
    * Pdt1=Pgmax
    * Pshedt1 = Pdt1 - pmax
    * pricet1 = cost of load shedding
    '''
    
    user_config.duals = True
    pmax=100
    Pdt1=211
    generators=[make_cheap_gen(pmax=pmax)]
    Pdt=[110,Pdt1,110]
    power_system, times = solve_problem(generators, do_reset_config=False,
        **make_loads_times(Pdt=Pdt))
    load=power_system.loads()[0]
    load_t1=load.power(times[1], evaluate=True)
    load_t1_shed=load.shed(times[1], evaluate=True)
    price_t1 = power_system.buses[0].price(times[1])
    assert load_t1==pmax 
    assert load_t1_shed==Pdt1-pmax
    assert price_t1 == user_config.cost_load_shedding


@istest
@with_setup(teardown=reset_config)
def reserve_fixed_amount():
    '''
    Create two generators, the cheaper one with a limit near the load.
    Require 50MW of reserve (which exceeds pmax of cheap)
    Ensure that: g2 is turned on to meet reserve req.

    Pg1+Pg2 = load
    Pavail1 + Pavail2 = load + reserve
    '''
    reserve_req = 30.0
    user_config.reserve_fixed = reserve_req
    
    generators=[
        make_cheap_gen(pmax=100), 
        make_expensive_gen(pmin=0, pmax=300, costcurveequation='5000+30P')]
    Pdt=[80, 90]    
        
    power_system, times = solve_problem(generators, 
        **make_loads_times(Pdt=Pdt))
    total_power = [sum([value(gen.power(time)) for gen in generators]) for time in times]
    total_available_power = [sum(value(gen.power_available(time)) 
        for gen in generators) for time in times]
        
#    print load_t1 + reserve_req
#    print [[value(gen.power(time)) for gen in generators] for time in times]
#    print total_available_power
#    print [reserve_req + Pd for Pd in Pdt]
    
    assert Pdt == total_power
    assert total_available_power >= [reserve_req + Pd for Pd in Pdt]


@istest
@with_setup(teardown=reset_config)
def startup_ramprate_default():
    '''
    Create two generators with a high limit,
    one limited by Pmax, the other by rampratemax.
    Ensure that the second gen. is started up to its limit to minimize shed.
    '''
    
    pmax = 100
    Pdt1 = 210
    generators = [
        make_cheap_gen(pmax=pmax),
        make_expensive_gen(rampratemax=100, 
            rampratemin=-200, pmin=95, pmax=200)
        ]
    initial = [{'power': 90}, {'status': 0}]
    Pdt = [90, Pdt1, 90]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt))

    # the expensive unit turns on
    assert_series_equal(pd.Series([0.0, 1.0, 0.0], index=times.times),
        generators[1].values('status', times.times))
    
    # the minimum load was shed
    assert_series_equal(pd.Series([0.0, 100.0, 0.0], index=times.times),
        generators[1].values('power', times.times))



""" 
Integration testing. Runs solve.problem() on all directories 
in the tests subpackage directory. These tests include ED,OPF,UC
tests designed to ensure that inputs parse correctly, 
constraints are active, and vizualizations and solution
output work.

Chances are good that if you make major modifications 
to the Minpower code you will break something here. 
Just fix it before you push your change.
"""

import sys,os,glob,traceback,logging
from minpower import config
from minpower.config import user_config
from minpower.commonscripts import joindir,splitFilename

def wipeTestSlate(dir):
    patterns=['*.png','*.lp','commitment.csv','dispatch.csv','error.txt']
    for pat in patterns:
        for f in glob.glob(joindir(dir,pat)): os.remove(f)
def hasPyscript(dir): return glob.glob(joindir(dir,'*.py'))

def main(solver=config.user_config.solver):
    logging.basicConfig(level=logging.CRITICAL, format='%(levelname)s: %(message)s')
    from minpower import solve
    
    dirNm=splitFilename(__file__)[0]
    if dirNm=='': dirNm='.'
    excludeL=[]
    for fileNm in os.listdir(dirNm):
        if fileNm in excludeL: continue
        testDir = joindir(dirNm, fileNm)
        if not os.path.isdir(testDir): continue
        print 'testing: ',fileNm
        wipeTestSlate(testDir)
        fResults=open(joindir(testDir,'results.txt'),'w+')
        fError=open(joindir(testDir,'error.txt'),'w+')
        sys.stdout=fResults #switch output to results file 
        if hasPyscript(testDir):
            sys.stdout = sys.__stdout__ #switch back to standard outputting
            os.system('python {s}'.format(s=hasPyscript(testDir)[0]))
        else:
            try: 
                user_config.scenarios = 2
                solve.solve_problem(testDir)
                sys.stdout = sys.__stdout__ #switch back to standard outputting
                fError.close()
                os.remove(joindir(testDir,'error.txt'))
            except: #write the error to file
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_tb(exc_traceback, file=fError )
                traceback.print_exception(exc_type, exc_value, exc_traceback, file=fError)
                sys.stdout = sys.__stdout__ #switch back to standard outputting
                print '\t had error' #note that this dir produced error
    else: 
        sys.stdout = sys.__stdout__ #switch back to standard outputting

if __name__ == "__main__": main()



from minpower import solve,config
import sys

def test(resultsfile='out.txt',logfile='out.log'):
    sys.stdout = open(resultsfile, 'w+')
    solve.solve_problem(
        datadir='/home/adam/ercot-data/ercot-2010-test-system/',
        shell=True,
        problemfile=False,
        visualization=False,
        csv=False,
        solver='cplex',
        num_breakpoints=10,
        hours_commitment=24,
        hours_commitment_overlap=12,
        get_duals=False,
        logging_file=logfile,
        solution_file=False
        )

if __name__ == "__main__": 
    config.updown_formulation=sys.argv[1]
    filename=config.updown_formulation.replace(' ','-')
    test(resultsfile=filename+'.txt',logfile=filename+'.log')


import logging
logging.basicConfig(level=logging.ERROR)

import nose
import numpy as np
import pandas as pd
from minpower.tests.test_utils import istest
from minpower.tests.test_integration import run_case, assert_series_equal

mipgap = 0.0001


@istest
def run_basic_checks():
    '''
    solutions should all make sense (cost=0 and power=0 if status=0)
    '''
    run_case('stochastic_short_case', deterministic_solve=True)


@istest
def mock_stochastic():
    '''
    the solution to the stochastic problem with only two scenarios,
    both with values equal to the deterministic forecast and prob=0.5,
    should have a total cost equal (within solver tolerance) to
    the deterministic forecat solution total cost
    '''
    slnS = run_case('stochastic_mock_case',
        scenarios_directory='scenarios/')
    slnD = run_case('stochastic_mock_case',
        deterministic_solve=True)

    costS = slnS.observed_cost.sum().sum()
    costD = slnD.observed_cost.sum().sum()

    diffpercent = np.abs((costS - costD) / costD)
    assert(diffpercent < mipgap)


@istest
def standalone():
    try:
        import tables
    except ImportError:
        raise nose.SkipTest('standalone mode requires pytables')

    slnA = run_case('stochastic_short_case', deterministic_solve=True)
    slnB = run_case('stochastic_short_case', deterministic_solve=True, standalone=True)

    assert(slnA.observed_cost.sum().sum() == slnB.observed_cost.sum().sum())
    


    
@istest
def expected_cost_case():
    '''
    ensure that a under-forecast wind case has
    an expected cost > observed cost
    '''
    # this is a short and simple case with 4hrs, but make it into two UC days
    hrs = dict(hours_commitment=2, hours_overlap=0)
    slnD = run_case('expected_observed_cost', deterministic_solve=True, **hrs)
    
    assert(slnD.expected_cost.sum().sum() > slnD.observed_cost.sum().sum())


@istest
def designed_diff_case():
    '''
    ensure that a simple case designed to produce a more expensive 
    deterministic forecast solution has a cheaper perfect forecast
    '''
    # this is a short and simple case with 4hrs, but make it into two UC days
    hrs = dict(hours_commitment=2, hours_overlap=0)
    slnP = run_case('deterministic_perfect_difference', perfect_solve=True, **hrs)
    slnD = run_case('deterministic_perfect_difference', deterministic_solve=True, **hrs)
    
    assert (slnP.generators_power.sum(axis=1) - \
        slnD.generators_power.sum(axis=1) == 0).all()
    assert slnP.observed_cost.sum().sum() < slnD.observed_cost.sum().sum()
    
    
@istest
def stochastic_load_shedding():
    '''
    Create a sheddable wind generator and one scenario that requires load shedding.
    Ensure that wind is shed in one scenario but not the other.
    '''
    
    sln = run_case('stochastic_shedding_case')
    
    epower = sln.generators_power_scenarios
    load_sched = sln.loads[0].schedule
    
    assert((epower['s0'].sum(axis=1) < load_sched).all())
    assert((epower['s1'].sum(axis=1) == load_sched).all())
    
@istest
def stochastic_gen_shedding():
    '''
    Create a sheddable wind generator and one scenario that requires wind shedding.
    Ensure that wind is shed in one scenario but not the other.
    '''
    
    sln = run_case('stochastic_shedding_wind_case')
    
    
    epower = sln.generators_power_scenarios
    load_sched = sln.loads[0].schedule
    load_sched.index = epower.major_axis
    # check that generation meets the load for all scenarios
    assert(epower.sum(axis=2).eq(load_sched, axis=0).all().all())

    # get wind power scenarios
    scenario_values = sln.generators[0].scenario_values.drop(
        'probability', axis=2)
    scenario_values = scenario_values[
        scenario_values.items[0]].T.ix[:len(epower.major_axis) -1] \
        .set_index(epower.major_axis) \
        .rename(
            columns=dict(zip(range(len(sln.scenarios)), sln.scenarios)))
    # calculate wind shed per scenario
    wind_shed = scenario_values - epower.minor_xs('g0')
    
    assert_series_equal(
        pd.Series(dict(s0=False, s1=True)), 
        (wind_shed > 0).all())



'''Test the constraint behavior of the generators'''

import logging
logging.basicConfig( level=logging.CRITICAL, format='%(levelname)s: %(message)s')

import pandas as pd
from pandas.util.testing import assert_series_equal
from minpower.generators import Generator_nonControllable, Generator_Stochastic
from minpower.optimization import value

from test_utils import *


@istest
def power_maximum():
    '''
    Create two generators, one cheap and one expensive. 
    Give the cheap one a maximum power limit lower than the load.
    Ensure that the cheaper unit is at its limit.
    '''
    pmax=100
    Pd=221
    generators=[
        make_cheap_gen(pmax=pmax),
        make_expensive_gen()
        ]
    power_system,times=solve_problem(generators,**make_loads_times(Pd))
    assert generators[0].power(times[0]) == pmax
    
@istest
def power_minimum():
    '''
    Create two generators, one cheap and one expensive, status on for both.
    Give the expensive one a min power limit.
    Ensure that the exp. unit is at its limit.
    '''
    pmin=10
    Pd=221
    generators=[
        make_cheap_gen(),
        make_expensive_gen(pmin=pmin)
        ]
    power_system,times=solve_problem(generators,**make_loads_times(Pd))
    assert generators[1].power(times[0]) == pmin

@istest
def ramp_up():
    '''
    Create 2 gen, cheaper one with ramp up limit. 
    Create load that ramps up faster than limit, at t1.
    Ensure that cheaper unit is at its limit.
    '''
    ramp_limit=80
    generators=[
        make_cheap_gen(rampratemax=ramp_limit),
        make_expensive_gen()
        ]
    initial = [{'power':250},{'power':250}]
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[250,350]))
    print generators[1]._parent_problem()
    assert generators[0].power(times[1]) - generators[0].power(times[0]) == ramp_limit
    
@istest
def ramp_down():
    '''
    Create 2 gen, expensive one with ramp down limit. 
    Create load that ramps down faster than limit at t1.
    Ensure that exp. unit is at its limit.
    '''    
    ramp_limit= -80
    generators=[
        make_cheap_gen(),
        make_expensive_gen(rampratemin=ramp_limit)
        ]
    initial = [{'power':250},{'power':250}]
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[550,450]))
    ramp_rate = value(generators[1].power(times[1])) - value(generators[1].power(times[0]))
    assertAlmostEqual(ramp_rate, ramp_limit)

@istest
def ramp_up_initial():
    '''
    Create 2 gen, cheaper one with ramp up limit. 
    Create load that ramps up faster than limit, at t0.
    Ensure that cheaper unit is at its limit.
    '''
    ramp_limit=80
    generators=[
        make_cheap_gen(rampratemax=ramp_limit),
        make_expensive_gen()
        ]
    initial = [{'power':250},{'power':250}]
    
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[350,350]))
    ramp_rate = generators[0].power(times[0]) - generators[0].initial_power
    assert ramp_rate == ramp_limit


@istest
def ramp_down_initial():
    '''
    Create 2 gen, expensive one with ramp down limit. 
    Create load that ramps down faster than limit, at t0.
    Ensure that exp. unit is at its limit.
    '''
    ramp_limit= -80
    generators = [
        make_cheap_gen(),
        make_expensive_gen(rampratemin=ramp_limit)]
    initial = [{'power':250},{'power':250}]
    
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[300,300]))
    ramp_rate = generators[1].power(times[0]) - generators[1].initial_power
    assert ramp_rate == ramp_limit


@istest
def cold_ramp_up():
    '''
    Create three generators, one that must ramp up from cold. 
    Ensure that the generator ramps at within its limit.
    '''
    ramp_limit_SU = 20
    ramp_limit = 30
    generators = [
        make_cheap_gen(pmax=200),
        make_mid_gen(pmin=10, rampratemax=ramp_limit, 
            startupramplimit=ramp_limit_SU),
        make_expensive_gen(pmin=1)]
    initial = [{'power': 200}, {'status': 0}, {'status': 0}]

    power_system, times = solve_problem(generators, gen_init=initial, 
        **make_loads_times(Pdt=[200, 230, 280]))
    
    ramp_rate_SU = value(generators[1].power(times[1])) - value(generators[1].power(times[0]))
    assert ramp_rate_SU <= ramp_limit_SU
    ramp_rate = value(generators[1].power(times[2])) - value(generators[1].power(times[1]))
    assert ramp_rate <= ramp_limit
    

@istest
def hot_shut_down():
    '''
    Create three generators, one that must shut down from hot. 
    Ensure that the generator ramps within its limit.
    '''
    ramp_limit_SD = -30
    ramp_limit = -20
    generators = [
        make_cheap_gen(pmax=200),
        make_expensive_gen(pmin=20, 
            rampratemin=ramp_limit, 
            shutdownramplimit=ramp_limit_SD
        ),
        make_mid_gen(pmin=1)]
    initial = [{'power': 200}, {'power': 50}, {'power': 50}]

    power_system, times = solve_problem(generators, gen_init=initial, 
        **make_loads_times(Pdt=[280, 150]))
        

    ramp_rate = value(generators[1].power(times[0])) - generators[1].initial_power
    assert ramp_limit <= ramp_rate

    ramp_rate_SD = value(generators[1].power(times[1])) - value(generators[1].power(times[0]))
    assert ramp_limit_SD == ramp_rate_SD 
    assert value(generators[1].status(times[1])) == 0     


@istest
def min_up_time():
    '''
    Create two generators: cheap with max power limit, exp. with min up time and min power limits.
    Create load that increases over the cheap limit at t1 and then reduces back to constant.
    Ensure that the expensive generator is on at t1 and t2, then turns off. 
    '''
    generators=[
        make_cheap_gen(pmax=100),
        make_expensive_gen(minuptime=2,pmin=5)   ]
    initial = [
        dict(power=80, status=True),
        dict(status=False)]
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[85,120,80,80]))
    limgen_status=[generators[1].status(t) for t in times]
    assert limgen_status == [0,1,1,0] or limgen_status == [1,1,0,0]



@istest
def min_down_time():
    '''
    Create three generators: 
    * cheap with max power limit
    * with min down time
    * expensive
    Create load that is higher than the cheap limit all but t1.
    Ensure that the mid generator is OFF at t1 and t2, then turns on again.
    Ensure that the hi generator is ON for t2 to make up the difference.
    '''
    generators=[
        make_cheap_gen(pmax=100),
        make_mid_gen(mindowntime=2,pmin=40),
        make_expensive_gen()]
    initial = [
        dict(power=100, status=True),
        dict(power=40, status=True,hoursinstatus=0),
        dict(status=False)]

    _, times = solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[150,10,140,140]))
    limgen_status = [generators[1].status(t) for t in times]
    expensive_status_t2 = generators[2].status(times[2])
    assert limgen_status==[1,0,0,1] and expensive_status_t2==1


@istest
def start_up_cost():
    '''
    Create two generators:
    * cheap with a max power
    * expensive with a start up cost.
    Create an increasing load.
    Ensure that the objective includes the startup cost.
    '''
    startupcost=9000
    generators=[
        make_cheap_gen(pmax=100),
        make_expensive_gen(pmin=10, startupcost=startupcost)   ]
    initial = [
        dict(power=80, status=True),
        dict(status=False)]
    _, times = solve_problem(generators,
        gen_init=initial,
        **make_loads_times(Pdt=[80,120]))
    assert generators[1].cost_startup(times[1]) == startupcost


@istest
def shut_down_cost():
    '''
    Create two generators, both on:
    * cheap with a max power
    * expensive with a start down cost and a min power.
    Create an decreasing load.
    Ensure that the objective includes the shutdown cost.
    '''
    shutdowncost=200
    generators=[
        make_cheap_gen(pmax=100),
        make_expensive_gen(shutdowncost=shutdowncost, pmin=20)   ]
    initial = [
        dict(power=80, status=True),
        dict(power=20, status=True)]
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[150,10]))
    assert generators[1].cost_shutdown(times[1])==shutdowncost


@istest
def min_up_time_longer():
    '''
    Create two generators: cheap with max power limit, exp. with min up time and min power limits.
    Create load that increases over the cheap limit at t1 and then reduces back to constant.
    Ensure that the expensive generator is on at t1 and t2, then turns off. 
    '''
    generators=[
        make_cheap_gen(pmax=100),
        make_expensive_gen(minuptime=8,pmin=5)   ]
    initial = [
        dict(power=80, status=True, hoursinstatus=1),
        dict(status=False, hoursinstatus=0)]
    _,times=solve_problem(generators,gen_init=initial,**make_loads_times(Pdt=[85,120,80,80,70,70,70,70,80,80]))
    limgen_status=[generators[1].status(t) for t in times]
    #logging.critical(limgen_status)
    #logging.critical([(generators[1].startup(t),generators[1].shutdown(t)) for t in times])
    #logging.critical(problem.constraints['minuptime_g1t01'])
    assert limgen_status == [0,1,1,1,1,1,1,1,1,0] or limgen_status == [1,1,1,1,1,1,1,1,0,0]
    
@istest
def wind_shedding():
    '''
    Create a sheddable wind generator and a situation that requires shedding.
    Ensure that the minimum power is shed.
    '''
    lts = make_loads_times(Pdt=[85, 110, 80, 80])
    Pwind = pd.Series([60, 100, 120, 60], index=lts['times'].strings.values)
    
    generators = [
        Generator_nonControllable(schedule=Pwind, sheddingallowed=True),
        make_expensive_gen()]
    
    power_system, times = solve_problem(generators, **lts)
    assert generators[0].power(times[2]) == 80
    assert generators[0].power(times[3]) == 60
    assert sum(generators[1].power(t).value for t in times) == 25 + 10 + 20
    
    
    # this should work for stochastic generators
    # TODO - a real stochastic test case
    #   this one just tests deterministic solve
    user_config.deterministic_solve = True
    generators = [
        Generator_Stochastic(schedule=Pwind, sheddingallowed=True),
        make_expensive_gen()]
    power_system, times = solve_problem(generators, **lts)
    assert generators[0].power(times[2]) == 80
    assert generators[0].power(times[3]) == 60
    assert sum(generators[1].power(t).value for t in times) == 25 + 10 + 20
    
    
@istest
def pmin_startup_limit():
    '''
    generators that have a pmin > rampratemax should be able to startup to pmin
    '''
    pmin = 185
    generators=[
        make_cheap_gen(pmin=0, pmax=100),
        make_mid_gen(pmin=pmin, pmax=200, rampratemax=10),
        make_expensive_gen(pmin=0, pmax=200)
        ]
    initial = [{'power': 100}, {'status': 0}, {'power': 85}]
    lts = make_loads_times(Pdt=[185, 210, 250, 285, 290])

    power_system, times = solve_problem(generators, gen_init=initial, **lts)
    
    power = generators[1].values('power')
    status = generators[1].values('status')
    assert((power[status==1] >= pmin).all())
    
        
@istest
def startupcost_can_shutdown():
    '''
    Generators with a startup cost must be able to shut down.
    '''
    pmax = 100
    startupcost = 100
    generators = [
        make_cheap_gen(pmax=pmax),
        make_expensive_gen(startupcost=startupcost, pmin=20)
        ]
    initial = [{'power': 90}, {'status': 0}]
    Pdt = [90, 120, 90]
    power_system, times = solve_problem(
        generators, gen_init=initial, **make_loads_times(Pdt=Pdt))

    # the expensive unit turns on and back off
    assert_series_equal(pd.Series([0.0, 1.0, 0.0], index=times.strings),
        generators[1].values('status'))
        
    assert_series_equal(pd.Series([0.0, startupcost, 0.0], index=times.strings),
        generators[1].values('startupcost'))
    



"""
    Solution models, output, and display for power system
    optimization problems. Matplotlib and networkx are used for
    visualization.
"""
import logging
import os
import pandas as pd
import numpy as np

try: 
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from commonscripts import (update_attributes, gen_time_dataframe, joindir,
    replace_all, getattrL, elementwiseAdd, writeCSV, transpose, within,
    set_trace)
from schedule import TimeIndex
from optimization import value
from config import user_config

for_publication = True

try:
    import matplotlib
    import matplotlib.pyplot as plot
    do_plotting = True
    # prettify plots
    plot.rc("xtick", direction="out")
    plot.rc("ytick", direction="out")
    plot.rc("ytick", labelsize='small')
    plot.rc("xtick", labelsize='small')
    plot.rc("legend", fancybox=True)
    if for_publication:
        plot.rc("font", size=16)
        plot.rc("font", family="serif")
except ImportError:
    logging.warning("Can't import matplotlib -- skipping plotting.")
    do_plotting = False


def full_filename(filename):
    if user_config.standalone:
        filename = '{}-{}'.format(user_config.pid, filename)
    return joindir(user_config.directory, filename)


def classify_problem(times, power_system):
    '''
    Classify the type of problem: ED, OPF, UC, or SCUC.
    and if it is stochastic or not.
    '''
    if not power_system.lines and len(times) == 1:
        kind = 'ED'
    elif len(times) == 1:
        kind = 'OPF'
    elif not power_system.lines:
        kind = 'UC'
    else:
        kind = 'SCUC'

    if power_system.is_stochastic:
        kind = 'stochastic_' + kind
    return kind


def make_solution(power_system, times, **kwargs):
    '''
    Create a :class:`solution.Solution` object for a power system over times.
    '''
    problem_type = dict(
        ED=Solution_ED,
        OPF=Solution_OPF,
        UC=Solution_UC,
        SCUC=Solution_SCUC,
        stochastic_UC=Solution_Stochastic_UC
    )
    kind = classify_problem(times, power_system)
    return problem_type[kind](power_system, times, **kwargs)


def make_multistage_solution(power_system, stage_times, stage_solutions):
    '''Create a multi stage solution'''
    if power_system.lines:
        logging.warning('no visualization for multistage SCUC yet')
    klass = MultistageStandalone if getattr(
        stage_solutions, 'path', False) else Solution_UC_multistage
    return klass(power_system, stage_times, stage_solutions)


class Solution(object):
    '''
    Solution information template for a power system over times.
    Each problem type has its own class for visualization and
    spreadsheet output, e.g. :class:`~solution.Solution_ED`.
    '''
    def __init__(self, power_system, times, datadir='.', is_stochastic=False):
        update_attributes(self, locals())
        self._resolved = False
        self._setup_powersystem()

        self._get_problem_info()
        self._get_outputs()
        self._get_costs()
        self._get_prices()

    def _setup_powersystem(self):
        '''shortcuts for power system attributes'''
        self.buses = self.power_system.buses
        self.lines = self.power_system.lines
        self.generators = self.power_system.generators()
        self.loads = self.power_system.loads()
        self.times_non_overlap = self.times.non_overlap()

    def get_values(self, items, method='power', time=None, evaluate=False):
        '''Get the attributes of all objects of a certain kind at a given time.'''
        if time:
            if evaluate:
                out = [value(getattr(
                    obj, method)(time, evaluate=True)) for obj in items]
            else:
                out = [value(getattr(obj, method)(time)) for obj in items]
        else:
            out = [value(getattr(obj, method)) for obj in items]
        return out

    def gen_time_df(self, method, non_overlap=True, evaluate=False,
                    generators=None):
        times = self.times_non_overlap if non_overlap else self.times
        if generators is None:
            generators = self.generators
        return gen_time_dataframe(generators, times,
            [self.get_values(generators, method, t, evaluate) for t in times])

    def _get_problem_info(self):
        self.solve_time = self.power_system.solution_time
        self.objective = float(value(self.power_system.objective))
        try:
            self.mipgap = self.power_system.mipgap
        except AttributeError:
            self.mipgap = None

    def _get_outputs(self):
        self.generators_power = self.gen_time_df('power')
        self.generators_status = _correct_status(self.gen_time_df('status'))

    def _get_costs(self):
        self.totalcost_generation = self.gen_time_df('cost', evaluate=True)
        self.fuelcost = self.gen_time_df('operatingcost', evaluate=True)
        self.fuelcost_true = self.gen_time_df('truecost').sum().sum()
        self.incremental_cost = self.gen_time_df('incrementalcost')

        times = self.times_non_overlap
        self.load_shed_timeseries = pd.Series(
            [sum(self.get_values(
                self.loads, 'shed', t, evaluate=True)) for t in times],
            index=times.strings.index)
        self.gen_shed_timeseries = self.gen_time_df('shed', evaluate=True,
            generators=self.power_system.get_generators_noncontrollable()
            ).sum(axis=1)

        self.load_shed = self.load_shed_timeseries.sum()
        self.gen_shed = self.gen_shed_timeseries.sum()

        if self.gen_shed > 0.01:
            logging.debug('generation shed: {}MW'.format(self.gen_shed))
        if self.load_shed > 0.01:
            logging.debug('load shed: {}MW'.format(self.load_shed))
        self._get_cost_error()

    def _get_cost_error(self):
        try:
            self.costerror = abs(self.fuelcost.sum(
            ).sum() - self.fuelcost_true) / self.fuelcost_true
        except ZeroDivisionError:
            self.costerror = 0

    def _get_prices(self):
        self.lmps = {}
        self.line_prices = {}
        for t in self.times:
            self.lmps[str(t)] = [bus.price(t) for bus in self.buses]
            self.line_prices[str(t)] = [line.price(t) for line in self.lines]

    def savevisualization(self, filename=None):
        '''Save the visualization to a file'''
        if filename is None:
            plot.show()
        else:
            plot.savefig(filename, bbox_inches='tight')
        plot.close()

    def show(self):
        '''Display the solution information to the terminal'''
        out = ['']
        out.extend(['Solution information', '-' * 10, ''])
        if len(self.times) < 5:
            out.append('Status')
            out.append(self.generators_status)
            out.append('Power')
            out.append(self.generators_power)
            out.append('IC')
            out.append(self.incremental_cost)
            out.append('')
            for t in self.times:
                if len(self.times) > 1:
                    out.append(t)
                out.extend(self.info_price(t))
                out.extend(self.info_loads(t))
                out.extend(self.info_buses(t))
                out.extend(self.info_lines(t))
                out.append('')

        out.extend(self.info_cost())
        print '\n'.join([str(o) for o in out])

    def info_status(self):
        return ['solved in {time:0.4f} sec'.format(time=self.solve_time)]

    def info_price(self, t):
        return ['price={}'.format(self.lmps[str(t)])]

    def info_loads(self, t):
        out = ['load info:']
        if len(self.buses) > 1:
            out.append('bus={}'.format(self.get_values(self.loads, 'bus')))
        out.append('name={}'.format(self.get_values(self.loads, 'name')))
        out.append('Pd={}'.format(self.get_values(self.loads, 'power', t)))
        return out

    def info_buses(self, t):
        buses = self.buses
        out = ['bus info:',
               'name={}'.format(getattrL(buses, 'name')),
               'Pinj={}'.format([bus.Pgen(t, evaluate=True) -
                                 bus.Pload(
                                     t, evaluate=True) for bus in buses]),
               'angle={}'.format(self.get_values(
                                 buses, 'angle', t) if len(buses) > 1 else []),
               'LMP={}'.format(self.lmps[str(t)])]
        return out

    def info_lines(self, t):
        lines = self.lines
        return ['line info:',
                'connecting={}'.format(
                    zip(getattrL(lines, 'frombus'), getattrL(lines, 'tobus'))),
                'Pk={}'.format(self.get_values(lines, 'power', t)),
                'price={}'.format(self.line_prices[str(t)])]

    def info_cost(self):
        return ['objective cost={}'.format(self.objective),
                'total cost of generation={}'.format(
                    self.totalcost_generation.sum().sum()),
                'linearized fuel cost of generation={}'.format(
                    self.fuelcost.sum().sum()),
                'polynomial fuel cost of generation={}'.format(
                    self.fuelcost_true),
                'percentage difference\t\t={diff:.2%}'.format(
                    diff=self.costerror),
                ]


class Solution_ED(Solution):
    def info_lines(self, t):
        return []

    def info_buses(self, t):
        return []

    def visualization(self, show_cost_also=False):
        ''' economic dispatch visualization of linearized incremental cost'''
        if not do_plotting:
            return
        t = self.times[0]
        price = self.lmps[str(t)][0]
        generators = self.generators
        loads = self.loads

        plotted_gens, names_gens, plotted_loads, names_loads = [], [], [], []
        minGen = min(getattrL(generators, 'pmin'))
        maxGen = max(getattrL(generators, 'pmax'))

        # save a plot of the price space - illustrating equal IC
        ax = plot.axes()
        for gen in generators:
            if gen.status(t):
                in_range, out_range = gen.bids.output_incremental_range()
                if gen.bids.is_pwl:
                    line, = plot.step(
                        in_range, out_range, where='pre', linestyle='-')
                else:
                    line, = plot.plot(in_range, out_range, linestyle='-', )
                plotted_gens.append(line)
                P = value(gen.power(t))
                IC = gen.incrementalcost(t)
                plot.plot(P, IC, '.', c=line.get_color(
                ), markersize=8, linewidth=2, alpha=0.7)
                names_gens.append(gen.name)
        if price is not None:
            grayColor = '.75'
            plot.plot([minGen, maxGen], [price, price], '--k', color=grayColor)
            plot.text(maxGen, price, '{p:0.2f} $/MWh'.format(
                p=price), color=grayColor, horizontalalignment='right')

        plot.xlabel('P [MWh]')
        if plotted_loads:
            plot.ylabel('Marginal Cost-Benifit [$/MWh]')
        else:
            plot.ylabel('Marginal Cost [$/MWh]')
        prettify_axes(ax)

        # plot.xlim(xmin=0,xmax=)
        ymin, _ = plot.ylim()
        if ymin < 0:
            plot.ylim(ymin=0)

        legendGens = plot.legend(plotted_gens, names_gens, fancybox=True,
                                 title='Generators:', loc='best')
        if plotted_loads:
            plot.legend(plotted_loads, names_loads,
                        fancybox=True, title='Loads:', loc='best')
            # add first legend to the axes manually bcs multiple legends get overwritten                        
            plot.gca().add_artist(legendGens)  

        self.savevisualization(filename=full_filename('dispatch.png'))

        if show_cost_also:
            # show a plot of the cost space, illustrating the linearization
            plot.figure()
            gensPlotted_price = plotted_gens
            plotted_gens, names_gens, plotted_loads, names_loads = [
            ], [], [], []
            for g, gen in enumerate(generators):
                if gen.status(t):
                    plotted_gens.append(gen.cost_model.plot(
                        P=value(gen.power(t)), 
                        linestyle='-', 
                        color=gensPlotted_price[g].get_color()
                        ))
                    names_gens.append(gen.name)
            for load in loads:
                if load.kind == 'bidding':
                    plotted_loads.append(load.bid(
                        t).plot(P=value(load.power(t)), linestyle=':'))
                    names_loads.append(load.name)
            plot.xlabel('P [MWh]')
            if plotted_loads:
                plot.ylabel('Cost-Benifit [$/h]')
            else:
                plot.ylabel('Cost [$/h]')
            legendGens = plot.legend(plotted_gens, names_gens, fancybox=True,
                                     title='Generators:', loc='best')
            if plotted_loads:
                plot.legend(plotted_loads, names_loads,
                            fancybox=True, title='Loads:', loc='best')
                plot.gca().add_artist(legendGens)  # add first legend to the axes manually bcs multiple legends get overwritten

            self.savevisualization(filename=full_filename('dispatch-cost.png'))

    def saveCSV(self):
        '''economic dispatch generator solution values in spreadsheet form'''
        def niceTF(value):
            return 0 if value == 0 else 1

        def nice_zeros(value):
            return 0 if value == 0 else value

        t = self.times.strings.index[0]

        out = pd.DataFrame(columns=['power', 'IC'],
                           index=self.generators_power.columns)
        out['power'] = self.generators_power.ix[t]
        out['IC'] = [nice_zeros(g.incrementalcost(self.times[0]))
                     for g in self.generators]

        if user_config.dispatch_decommit_allowed:
            # add status, but label all units at Pmin as OFF
            out['status'] = self.generators_status.ix[t]
            out = out.drop('IC', axis=1)

        out.index = [gen.name for gen in self.generators]
        out.index.name = 'name'

        out.to_csv(full_filename('dispatch.csv'))


class Solution_OPF(Solution):
    def visualization(self, filename=None):
        '''power flow visualization'''
        if not do_plotting:
            return

        try:
            import networkx as nx
        except ImportError:
            logging.warning("Could'nt import networkx -- skipping plotting.")
            return

        buses, lines, t = self.buses, self.lines, self.times[0]

        G = nx.DiGraph()
        for bus in buses:
            Pinj = value(bus.Pgen(t)) - value(bus.Pload(t))
            G.add_node(bus.name, Pinj=Pinj)
        for line in lines:
            P = value(line.power(t))
            if P >= 0:
                G.add_edge(line.frombus, line.tobus, P=P, Plim=line.pmax)
            else:
                G.add_edge(line.tobus, line.frombus, P=-P, Plim=-line.pmin)

        pos = nx.spectral_layout(G)
        Pinj = [ndata['Pinj'] for bus, ndata in G.nodes(data=True)
                if 'Pinj' in ndata]
        nx.draw(G, node_color=Pinj, pos=pos, node_size=1500,
                alpha=.7, cmap=plot.cm.get_cmap('RdYlBu'), fontsize=30)
        cb = plot.colorbar(shrink=.8)
        cb.set_label('injected power [MW]', fontsize=15)

        Plines = [edata['P'] for _, t, edata in G.edges(data=True)
                  if 'P' in edata]
        atLimLines = [(f, t) for f, t, edata in G.edges(
            data=True) if within(edata['P'], val=edata['Plim'], eps=1e-3)]
        nx.draw_networkx_edges(
            G, edge_color='0.6', pos=pos, width=Plines, alpha=0.5)
        nx.draw_networkx_edges(G, edgelist=atLimLines, edge_color='r',
                               pos=pos, width=Plines, alpha=0.5)
        self.savevisualization(full_filename('powerflow.png'))

    def saveCSV(self):
        '''OPF generator and line power values in spreadsheet form'''
        t = self.times[0]
        gens = self.generators
        lines = self.lines

        fields, data = [], []
        fields.append('generator name')
        data.append(self.get_values(gens, 'name'))
        fields.append('u')
        data.append(self.get_values(gens, 'status', t))
        fields.append('P')
        data.append(self.get_values(gens, 'power', t))
        fields.append(
            'IC')
        data.append(self.get_values(gens, 'incrementalcost', t))

        writeCSV(fields, transpose(
            data), filename=full_filename('powerflow-generators.csv'))

        fields, data = [], []
        fields.append('from')
        data.append(self.get_values(lines, 'frombus'))
        fields.append('to')
        data.append(self.get_values(lines, 'tobus'))
        fields.append('power')
        data.append(self.get_values(lines, 'power', t))
        fields.append(
            'congestion shadow price')
        data.append(self.line_prices[str(t)])

        writeCSV(fields, transpose(
            data), filename=full_filename('powerflow-lines.csv'))

    def info_price(self, t):
        return []  # built into bus info


class Solution_UC(Solution):
    def info_lines(self, t):
        return []

    def info_buses(self, t):
        return []

    def saveCSV(self, save_final_status=False):
        '''generator power values and statuses for unit commitment'''

        # mutlistage UC
        # for stochastic: output is the generators' power under observed wind
        self.generators_power.to_csv(full_filename('commitment-power.csv'))
        self.generators_status.to_csv(full_filename('commitment-status.csv'))

    def visualization(self, withPrices=True):
        '''generator output visualization for unit commitment'''
        if not do_plotting:
            return

        prices = [self.lmps[str(t)][0] for t in self.times]
        stack_plot_UC(
            self, self.power_system.generators(), self.times, prices, withPrices=withPrices)
        self.savevisualization(full_filename('commitment.png'))


class Solution_SCUC(Solution_UC):
    def visualization(self):
        logging.warning('no visualization for SCUC. Spreadsheet output is valid, except for the price column is the price on first bus only.')


class Solution_UC_multistage(Solution_UC):
    '''
    Muti-stage unit commitment. Each stage represents one optimization problem.
    Each element of the list :param:stage_solutions is a :class:`~results.Solution_UC` object.

    '''
    def __init__(self, power_system, stage_times, stage_solutions):
        update_attributes(self, locals(),
                          exclude=['stage_solutions', 'stage_times'])
        self._resolved = power_system.is_stochastic \
            or user_config.deterministic_solve or user_config.perfect_solve

        self.is_stochastic = any(sln.is_stochastic for sln in stage_solutions)

        times = pd.concat(
            [times.non_overlap().strings for times in stage_times]).index
        self.times = TimeIndex(times)
        self.times.set_initial(stage_times[0].initialTime)

        self.objective = self._sum_over('objective', stage_solutions)
        self.solve_time = self._sum_over('solve_time', stage_solutions)
        self.mipgaps = pd.Series([sln.mipgap for sln in stage_solutions])

        self._get_outputs(stage_solutions)
        self._get_costs(stage_solutions)
        self._get_prices(stage_solutions)

    def _sum_over(self, attrib, stage_solutions):
        return sum(getattr(sln, attrib) for sln in stage_solutions)

    def _concat(self, attrib, slns):
        return pd.concat([getattr(sln, attrib) for sln in slns])

    def _get_outputs(self, slns):
        '''the outputs under observed wind'''
        self.generators_power = self._concat('generators_power', slns)
        self.generators_status = self._concat('generators_status', slns)
        if self._resolved:
            self.expected_power = self._concat('expected_power', slns)
            self.expected_status = self._concat('expected_status', slns)

    def _get_costs(self, slns):
        self.expected_cost = self.totalcost_generation = \
            self._concat('expected_totalcost'
                         if self._resolved else 'totalcost_generation', slns)
        self.load_shed_timeseries = self._concat('load_shed_timeseries', slns)
        self.gen_shed_timeseries = self._concat('gen_shed_timeseries', slns)

        self.load_shed = self.load_shed_timeseries.sum()
        self.gen_shed = self.gen_shed_timeseries.sum()

        if self._resolved:
            self.observed_cost = self.totalcost_generation = \
                self._concat('observed_totalcost', slns)

    def info_cost(self):
        resolved = self._resolved
        expected = 'expected ' if resolved else ''
        observed = 'observed ' if resolved else ''
        out = []
        out.append('total {}generation cost = {}'.format(
            expected, self.expected_cost.sum().sum()))
        if resolved:
            out.append('total {}generation cost = {}'.format(
                observed, self.observed_cost.sum().sum()))

        return out

    def _get_prices(self, stage_solutions):
        self.lmps = {}
        try:
            for stage in stage_solutions:
                self.lmps.update(stage.lmps)
        except:
            # no lmps in stochastic solutions right now
            pass

    def show(self):
        out = []
        out.extend(self.info_status())
        out.extend(self.info_cost())
        out.extend(self.info_shedding())
        print '\n'.join(out)

    def info_generators(self):
        return []

    def info_loads(self):
        return []

    def info_status(self):
        return ['solved multistage problem in a total solver time ' +
                'of {time:0.4f} sec'.format(time=self.solve_time)]

    def info_shedding(self):
        return [
            'total load shed={}MW'.format(self.load_shed)
            if self.load_shed > 0.01 else '',
            'total gen shed={}MW'.format(self.gen_shed)
            if self.gen_shed > 0.01 else '',
        ]


class MultistageStandalone(Solution_UC_multistage):
    def __init__(self, power_system, stage_times, store):
        self.power_system = power_system
        self.is_stochastic = power_system.is_stochastic
        self._resolved = self.is_stochastic or \
            user_config.deterministic_solve or user_config.perfect_solve
        times = pd.concat(
            [times.non_overlap().strings for times in stage_times]).index
        self.times = TimeIndex(times)
        self.times.set_initial(stage_times[0].initialTime)

        self.expected_cost = self.totalcost_generation = store['expected_cost']

        if self._resolved:
            self.observed_cost = self.totalcost_generation = \
                store['observed_cost']

        self.generators_power = store['power']
        self.generators_status = store['status']
        self.load_shed_timeseries = store['load_shed']
        self.gen_shed_timeseries = store['gen_shed']

        self.load_shed = store['load_shed'].sum()
        self.gen_shed = store['gen_shed'].sum()

        self.solve_time = store['solve_time'].sum()


class Solution_Stochastic(Solution):
    def __init__(self, power_system, times, datadir='.', is_stochastic=True):
        update_attributes(self, locals())
        self._setup_powersystem()

        self.scenarios = sorted(self.power_system._scenario_instances.keys())
        self.stage_start = times.strings.index[0]

        gen = self.power_system.get_generator_with_scenarios()

        try:  # single stage setup
            self.probability = gen.scenario_values.probability.dropna()
        except AttributeError:  # staged setup
            self.probability = gen.scenario_values[
                self.stage_start].probability.dropna()

        self.probability.index = self.scenarios

        self._get_problem_info()
        self._get_outputs()
        self._get_costs()
        self._get_prices()

    def stg_panel(self, method, 
        generators=None,
        no_overlap=True,
        evaluate=False):
        if generators is None:
            generators = self.generators
            
        times = self.times_non_overlap if no_overlap else self.times
        if evaluate:
            getval = lambda s, gen, t: value(
                getattr(gen, method)(t, s, evaluate=True))
        else:
            getval = lambda s, gen, t: value(getattr(gen, method)(t, s))
        return pd.Panel(
            [[[getval(s, gen, t) for gen in generators]
                for t in times] for s in self.scenarios],
            items=[s for s in self.scenarios],
            major_axis=times.strings.index,
            minor_axis=[str(g) for g in generators])

    def gen_time_df(self, method, scenario, non_overlap=True, evaluate=False,
                    generators=None):
        if generators is None:
            generators = self.generators

        if evaluate:
            getval = lambda gen, t: value(
                getattr(gen, method)(t, scenario, evaluate=True))
        else:
            getval = lambda gen, t: value(getattr(gen, method)(t, scenario))

        times = self.times_non_overlap if non_overlap else self.times
        return gen_time_dataframe(generators, times,
                                  [[getval(gen, t) for t in times] for gen in generators])

    def _get_outputs(self, resolve=False):
        if resolve:
            # observed generator power
            # resolved on the first scenario instance
            # -- no more scenario labeling is needed
            self.generators_power = self.gen_time_df('power', None)
            self.generators_status = self.gen_time_df('status', None)

        else:
            self.generators_power_scenarios = self.stg_panel('power')
            self.generators_status_scenarios = \
                _correct_status(self.stg_panel('status'))
            self.expected_status = \
                self.generators_status_scenarios[self.scenarios[0]]
            self.generators_status = self.expected_status.copy()
            self.expected_power = self.generators_power = self._calc_expected(
                self.generators_power_scenarios)
        return

    def _calc_expected(self, panel):
        out = panel.copy()
        for s, pr in self.probability.iteritems():
            out[s] *= pr
        return out.sum(axis=0)

    def _calc_expected_cost(self, method):
        '''
        calculate expected cost by getting costs over all scenarios,
        multiplying each scenario cost by its probability,
        and summing over the scenarios
        '''
        return self._calc_expected(self.stg_panel(method, evaluate=True))

    def _get_costs(self, resolve=False):
        if resolve:
            # resolved on the first scenario instance
            # -- no more scenario labeling is needed
            s = None
            self.observed_totalcost = self.totalcost_generation = \
                self.gen_time_df('cost', s, evaluate=True)
            self.observed_fuelcost =  self.fuelcost = \
                self.gen_time_df('operatingcost', s, evaluate=True)

            # calc observed load shed
            if len(self.loads) > 1:
                raise NotImplementedError
            self.load_shed_timeseries = pd.Series(
                [self.power_system.loads()[0].shed(time, evaluate=True)
                    for time in self.times_non_overlap],
                index=self.generators_power.index)

            self.gen_shed_timeseries = self.gen_time_df('shed', s,
                                                        evaluate=True,
                                                        generators=self.power_system.get_generators_noncontrollable()
                                                        ).sum(axis=1)

            self.load_shed = self.load_shed_timeseries.sum()
            self.gen_shed = self.gen_shed_timeseries.sum()

            if self.gen_shed > 0.01:
                logging.debug('generation shed: {}MW'.format(self.gen_shed))
            if self.load_shed > 0.01:
                logging.debug('load shed: {}MW'.format(self.load_shed))

        else:
            # get expected cost of non_overlap times
            self.expected_totalcost = self._calc_expected_cost('cost')
            self.expected_fuelcost = self._calc_expected_cost('operatingcost')

            # calculate expected load, gen shed
            self.expected_gen_shed_timeseries = self._calc_expected(self.stg_panel('shed', evaluate=True, 
                generators=self.power_system.get_generators_noncontrollable()))            
            self.expected_gen_shed = self.expected_gen_shed_timeseries.sum().sum()
                
            self.expected_load_shed_timeseries = self._calc_expected(
                self.stg_panel('shed', evaluate=True, 
                    generators=self.power_system.loads()))
            self.expected_load_shed = self.expected_load_shed_timeseries.sum()

            if self.expected_gen_shed > 0.01:
                logging.debug('expected generation shed: {}MW'.format(
                    self.expected_gen_shed))
            if self.expected_load_shed > 0.01:
                logging.debug('expected load shed: {}MW'.format(
                    self.expected_load_shed))
            
                

    def _get_cost_error(self):
        pass

    def _get_prices(self):
        pass

    def info_cost(self):
        return ['expected cost= {}'.format(self.expected_totalcost.sum().sum())]

    def show(self):
        '''Display the solution information to the terminal'''
        # self.problem.scenario_tree.pprintSolution()
        out = ['']
        out.extend(['Solution information', '-' * 20])
        out.extend(self.info_cost())
        print '\n'.join(out)


class Solution_Stochastic_UC(Solution_Stochastic):
    def saveCSV(self):
        '''generator power and status values for stochastic unit commitment'''
        try:
            self.generators_power.to_csv(full_filename('commitment-power.csv'))
        except AttributeError:
            logging.warn('''didnt resolve - no power
                 values are available for csv format''')
        self.generators_status.to_csv(full_filename('commitment-status.csv'))


def _correct_status(status):
    # correct for strange solver values returned on resolve
    status[status > 0.99] = 1
    status[status < 0.01] = 0
    return status.astype(int)


def _colormap(numcolors, colormapName='gist_rainbow', mincolor=1):
    cm = matplotlib.cm.get_cmap(colormapName)
    return [cm(1. * i / numcolors) for i in range(mincolor, numcolors + mincolor)]


def stack_plot_UC(solution, generators, times, prices,
                  withPrices=True,
                  seperate_legend=False,
                  hours_tick_interval=None
                  ):
    withPrices = withPrices and any(prices)
    font_big = {'fontsize': 15}
    figWidth = .85
    figLeft = (1 - figWidth) / 2
    yLabel_pos = {'x': -0.12, 'y': 0.5}

    plot.figure(figsize=(10, 4), dpi=180)
    ax = plot.axes([figLeft, .1, figWidth, .6])
    ax.set_ylabel('energy [MWh]', ha='center', **font_big)
    ax.yaxis.set_label_coords(**yLabel_pos)
    prettify_axes(ax)

    T = times.times.values
    bar_width = times.intervalhrs / 24.0  # maplotlib dates have base of 1day

    convert_to_GW = solution.generators_power.max().sum() > 20e3
    
    power = solution.generators_power.copy()
    if convert_to_GW:
        power = power / 1e3

    colors = _colormap(len(generators), colormapName='Blues')

    power.plot(ax=ax, kind='bar', legend=True, stacked=True, color=colors, edgecolor='none')
    
    # show prices
    if withPrices:
        prices = pd.Series(prices, index=times.times)
        prices[prices == user_config.cost_load_shedding] = None
        prices = prices.dropna()
        if len(prices):
            axes_price = plot.axes([figLeft, .75, figWidth, .2], sharex=ax)

            # HACK - pandas uses this index for the bar plots
            # and since they share an x axis, use the same x values here            
            pd.Series(prices.values, index=np.arange(len(prices)) + 0.25).plot(
                ax=axes_price, drawstyle='steps')
            
            # axes_price.step(times.times.tolist() + [times.End], prices.values.tolist() + [prices[prices.index[-1]]],
            #                 where='post')  # start from 1 past initial time
            axes_price.set_ylabel('price\n[$/MWh]', ha='center', **font_big)
            axes_price.yaxis.set_label_coords(**yLabel_pos)
            plot.setp(axes_price.get_xticklabels(), visible=False)
            # format the price axis nicely
            axes_price.set_ylim(.9 * prices.min(), 1.1 * prices.max())
            axes_price.yaxis.set_major_locator(
                matplotlib.ticker.MaxNLocator(5))
            prettify_axes(axes_price)

    ax.xaxis_date()

    for label in ax.get_xticklabels():
        label.set_ha('right')
        label.set_rotation(30)

    # format the time axis nicely
    if hours_tick_interval is None:
        if len(times) < 10:
            pass
        else:
            if 24 * 10 > times.spanhrs > 48:
                # TODO - this isn't working because pandas bar internals
                # sets the x values to floats, instead of the actual times
                # https://github.com/pydata/pandas/blob/master/pandas/tools/plotting.py#L1257
                ax.xaxis.set_major_locator(matplotlib.dates.DayLocator())
                ax.xaxis.set_major_formatter(
                    matplotlib.dates.DateFormatter('%Y-%m-%d'))
                ax.xaxis.set_minor_locator(matplotlib.dates.HourLocator())
            elif times.spanhrs < 48:
                ax.xaxis.set_major_locator(matplotlib.dates.HourLocator())
                ax.xaxis.set_major_formatter(
                    matplotlib.dates.DateFormatter('%H:%M'))
            # otherwise use defaults
    else:
        ax.xaxis.set_major_locator(
            matplotlib.dates.HourLocator(interval=hours_tick_interval))
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%H:%M'))

    # format the power axis nicely
    if convert_to_GW:
        ax.set_ylabel('energy [GWh]', ha='center', **font_big)
    ax.autoscale_view()

def prettify_axes(ax):
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()


def shrink_axis(ax, percent_horizontal=0.20, percent_vertical=0):
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * (1 -
                    percent_horizontal), box.height * (1 - percent_vertical)])



from minpower.tests.test_utils import (user_config, istest, 
    with_setup, reset_config)
from minpower.get_data import parsedir
import os


basedir = os.path.split(__file__)[0]

@istest
@with_setup(teardown=reset_config)
def check_pmin():
    user_config.directory = os.path.join(basedir, 'uc')
    generators, loads, _, times, _, data = parsedir()
    
    assert(data['generators'].pmin.tolist() == \
        map(lambda gen: gen.pmin, generators))


'''Test the constraint behavior of an OPF'''
from minpower import powersystems
from minpower.optimization import value
from minpower.commonscripts import Series
from minpower.config import user_config

from test_utils import (istest, get_duals, with_setup, reset_config,
    make_cheap_gen, make_mid_gen, make_expensive_gen, 
    singletime, make_loads_times, 
    solve_problem)

def test_config():
    assert(user_config.duals == False)
    
@istest
@with_setup(get_duals, reset_config)
def line_limit_high():
    '''
    Create a two bus system
        -cheap gen on busA
        -expensive gen and load on busB, 
        -line limit lower than load
    Ensure that: 
        - the line is at its limit 
        - the congestion price is equal to the diff. in LMPs
    '''
    pmax=100
    generators=[
        make_cheap_gen(bus='A'),
        make_expensive_gen(bus='B')
    ]
    
    lines=[powersystems.Line(pmax=pmax, frombus='A', tobus='B')]
    power_system, times = solve_problem(generators, do_reset_config=False,
        lines=lines, **make_loads_times(Pd=225,bus='B'))
    Pline = value(lines[0].power(times[0]))
    lmps = [b.price(times[0]) for b in power_system.buses]
    congestion_price = lines[0].price(times[0])
    assert Pline == pmax and congestion_price == (lmps[1] - lmps[0])

@istest
@with_setup(get_duals, reset_config)
def line_limit_low():
    '''
    Create a two bus system
        -cheap gen on busA
        -expensive gen and load on busB, 
        -line (from B to A) limit lower than load
    Ensure that: 
        - the line is at its limit 
        - the congestion price is equal to the diff. in LMPs
    '''
    pmin = -100
    generators=[
        make_cheap_gen(bus='A'),
        make_expensive_gen(bus='B')
    ]    
    lines=[powersystems.Line(pmin=pmin, frombus='B', tobus='A')]
    power_system,times=solve_problem(generators, do_reset_config=False,
        lines=lines, **make_loads_times(Pd=225,bus='B'))
    Pline = lines[0].power(times[0])
    lmps=[b.price(times[0]) for b in power_system.buses]
    congestion_price = lines[0].price(times[0])
    assert Pline==pmin and congestion_price == -1*(lmps[1] - lmps[0])


@istest
@with_setup(get_duals, reset_config)
def three_buses():
    '''
    Create a three bus system
        -cheap gen and load on busA
        -mid gen and load on busB
        -expensive gen and load on busC
        -three lines connecting all three buses (triangle)
        -create congestion by setting a line limit
    Ensure that: 
        - the total load is met
        - the lmps are different
    '''
    pmax=50
    Pd=[105,225,302]
    generators=[
        make_cheap_gen(bus='A'),
        make_mid_gen(bus='B'),
        make_expensive_gen(bus='C'),
        ]    
    loads=[
        powersystems.Load(schedule=Series(Pd[0],singletime), bus='A'),
        powersystems.Load(schedule=Series(Pd[1],singletime), bus='B'),
        powersystems.Load(schedule=Series(Pd[2],singletime), bus='C')
        ]
    lines=[
        powersystems.Line(frombus='A', tobus='B'),
        powersystems.Line(frombus='A', tobus='C', pmax=pmax),
        powersystems.Line(frombus='B', tobus='C', pmax=pmax),
        ]
    power_system,times=solve_problem(generators, do_reset_config=False,
        times=singletime,loads=loads,lines=lines)
    num_lmps=len(set(b.price(times[0]) for b in power_system.buses))
    total_load = value(sum(b.Pload(times[0]) for b in power_system.buses))
    assert total_load==sum(Pd) and num_lmps>1

def test_config_cleared():
    assert(user_config.duals == False)



"""
Stochastic scenario models for schedules.
"""
from coopr.pyomo import AbstractModel, Set, Param, Boolean, Var
from coopr.pysp.scenariotree import ScenarioTree
from coopr.pysp.ef import create_ef_instance
import gc
import logging

def construct_simple_scenario_tree(power_system, times, time_stage=None):
    '''Construct a simple scenario tree instance'''

    gen = power_system.get_generator_with_scenarios()    
    probabilities = gen.scenario_values[times.Start].probability.dropna().values.tolist()

    tree = new_scenario_tree_model()

    prob_set = range(len(probabilities))
    if time_stage is None:
        scenario_names = ['s{n}'.format(n=n) for n in prob_set]
        node_names = ['n{n}'.format(n=n) for n in prob_set]
    else:
        scenario_names = ['s{n}t{t}'.format(n=n, t=time_stage)
                          for n in prob_set]
        node_names = ['n{n}t{t}'.format(n=n, t=time_stage) for n in prob_set]

    tree.Stages.add('first stage', 'second stage')
    tree.Nodes.add('root')
    for nm in node_names:
        tree.Nodes.add(nm)
    tree.NodeStage['root'] = 'first stage'
    for nm in node_names:
        tree.NodeStage[nm] = 'second stage'
    tree.Children['root'] = node_names

    tree.ConditionalProbability['root'] = 1.0
    for i, pr in enumerate(probabilities):
        tree.ConditionalProbability[node_names[i]] = pr

    for nm in scenario_names:
        tree.Scenarios.add(nm)
    for i, nm in enumerate(scenario_names):
        tree.ScenarioLeafNode[nm] = node_names[i]
    # tree.Scenarios.pprint()

    power_system._scenario_tree_instance = tree
    logging.debug('constructed tree for stage {s}'.format(s=time_stage))
    # note that stage variables still need definition
    # as do the values of the per scenario variables
    return


def define_stage_variables(power_system, times):
    # scenario_tree.Stages.pprint()
    # scenario_tree.StageVariables.pprint()

    # create sets of variable names (not actual variables) for each stage
    variables_first_stage = Set()
    variables_second_stage = Set()

    for gen in power_system.get_generators_without_scenarios():
        # for each non stochastic generator
        if gen.is_controllable:
            variables_first_stage.add(
                str(gen.get_variable('status', indexed=True, time=None)) + '[*]')
            variables_second_stage.add(
                str(gen.get_variable('power', indexed=True, time=None)) + '[*]')
        # note - appending '[*]' to the indicies is required to get
        # pysp to assign all the variables in the array to a shape

    if power_system.shedding_mode:
        for gen in filter(lambda gen: gen.shedding_mode, 
            power_system.get_generators_noncontrollable()):
            variables_second_stage.add(
                str(gen.get_variable('power_used', time=None, indexed=True)) + '[*]')
        for load in power_system.loads():
            variables_second_stage.add(
                str(load.get_variable('power', time=None, indexed=True)) + '[*]')
    
    # variables_first_stage.pprint()
    scenario_tree = power_system._scenario_tree_instance
    
    scenario_tree.StageVariables['first stage'] = variables_first_stage
    scenario_tree.StageVariables['second stage'] = variables_second_stage

    scenario_tree.StageCostVariable['first stage'] = str(
        power_system.cost_first_stage())
    scenario_tree.StageCostVariable['second stage'] = str(
        power_system.cost_second_stage())

def create_problem_with_scenarios(power_system, times):
    logging.debug('constructing scenario tree')

    scenario_tree = ScenarioTree(
        scenarioinstance=power_system._model,
        scenariotreeinstance=power_system._scenario_tree_instance)

    if not scenario_tree.validate():
        for s, scenario in enumerate(scenario_tree._scenarios):
            print s, scenario
        raise ValueError('not a valid scenario tree')

    gen = power_system.get_generator_with_scenarios()

    scenario_instances = {}

    # construct scenario instances
    logging.debug('constructing scenario instances')
    gc.disable()
    for s, scenario in enumerate(scenario_tree._scenarios):
        scenario_instance = power_system._model.clone()

        power = getattr(scenario_instance, 'power_{}'.format(str(gen)))
        # set the values of the parameter for this scenario
        logging.debug('setting scenario values for s%i' % s)
        scenario_vals = gen._get_scenario_values(times, s=s)
        for t, time in enumerate(times):
            power[time] = scenario_vals[t]

        # power.pprint()
        scenario_instance.preprocess()
        scenario_instances[scenario._name] = scenario_instance

    gc.enable()
    scenario_tree.defineVariableIndexSets(power_system._model)
    full_problem_instance = create_ef_instance(
        scenario_tree, scenario_instances)
#        generate_weighted_cvar = generate_weighted_cvar,
#        cvar_weight = cvar_weight,
#        risk_alpha = risk_alpha,
#        cc_indicator_var_name = cc_indicator_var_name,
#        cc_alpha = cc_alpha)
        
    # could generate cvar here
    # full_problem_instance.pprint()

    # relax the non-anticipatory constraints on the generator status variables
    # beyond the UC time horizon
    if len(times.post_horizon()) > 0:
        logging.debug('removing non-anticipatory constraints for post-commitment horizon')
    for time in times.post_horizon():
        for scenario in scenario_instances.keys():
            for gen in power_system.generators():
                if not gen.is_controllable:
                    continue
                u = gen.status().name
                delattr(full_problem_instance, 
                    '{s}_{u}_{t}'.format(s=scenario, u=u, t=str(time)))

    power_system.stochastic_formulation = True
    power_system._stochastic_instance = full_problem_instance
    power_system._scenario_tree = scenario_tree
    power_system._scenario_instances = scenario_instances
    return


def get_scenario_based_costs(scenario_tree, scenario_instances):
    # scenario_tree.pprintCosts(scenario_instances)
    costs = dict()
    for node in scenario_tree._tree_node_map.values():
        scenarios = node._scenarios
        if len(scenarios) == 1:
            costs[node._scenarios[0]._name] = \
                node.computeExpectedNodeCost(scenario_instances)
        else:
            continue
    return costs


def update_variables(power_system, times):
    '''Convert all variables into dictionaries of their solved values, keyed by scenario'''
    variable_names = power_system._model.active_components(Var).keys()
    values = dict([(nm, {}) for nm in variable_names])
    for s, scenario in power_system._scenario_instances.items():
        for var_name, var in scenario.active_components(Var).items():
            if var.is_indexed():
                    values[var_name][s] = dict(
                        [(idx, var_val.value) for idx, var_val in var.iteritems()])
            else:
                    values[var_name][s] = var.value
    power_system._per_scenario_values = values


def new_scenario_tree_model():
    # https://software.sandia.gov/trac/coopr/browser/coopr.pysp/trunk/coopr/pysp/util/scenariomodels.py
    scenario_tree_model = AbstractModel()

    # all set/parameter values are strings, representing the names of various
    # entities/variables.

    scenario_tree_model.Stages = Set(ordered=True)
    scenario_tree_model.Nodes = Set()

    scenario_tree_model.NodeStage = Param(scenario_tree_model.Nodes,
                                          within=scenario_tree_model.Stages)
    scenario_tree_model.Children = Set(scenario_tree_model.Nodes,
                                       within=scenario_tree_model.Nodes, ordered=True)
    scenario_tree_model.ConditionalProbability = Param(
        scenario_tree_model.Nodes)

    scenario_tree_model.Scenarios = Set(ordered=True)
    scenario_tree_model.ScenarioLeafNode = Param(scenario_tree_model.Scenarios,
                                                 within=scenario_tree_model.Nodes)

    scenario_tree_model.StageVariables = Set(scenario_tree_model.Stages)
    scenario_tree_model.StageCostVariable = Param(scenario_tree_model.Stages)

    # scenario data can be populated in one of two ways. the first is "scenario-based",
    # in which a single .dat file contains all of the data for each scenario. the .dat
    # file prefix must correspond to the scenario name. the second is "node-based",
    # in which a single .dat file contains only the data for each node in the scenario
    # tree. the node-based method is more compact, but the scenario-based method is
    # often more natural when parameter data is generated via simulation. the default
    # is scenario-based.
    scenario_tree_model.ScenarioBasedData = Param(within=Boolean,
                                                  default=True, mutable=True)

    # do we bundle, and if so, how?
    scenario_tree_model.Bundling = Param(within=Boolean,
                                         default=False, mutable=True)
    scenario_tree_model.Bundles = Set()  # bundle names
    scenario_tree_model.BundleScenarios = Set(scenario_tree_model.Bundles)

    return scenario_tree_model



from test_utils import *
from minpower.results import make_solution

@istest
@with_setup(teardown=reset_config)
def check_power_status():
    '''
    Ensure that the solution generator_status and generator_power attributes
    match with the optimization values returned.
    '''
    generators=[
        make_cheap_gen(pmax=100),
        make_mid_gen(pmax=20),
        make_expensive_gen()
    ]

    user_config.duals = True
    power_system, times = solve_problem(generators, 
        **make_loads_times(Pdt=[80,110,130]))
    
    sln = make_solution(power_system, times)
    t0 = sln.generators_power.index[0]
    
    assert(sln.generators_power.ix[t0, 'g0'] == \
        value(generators[0].power(times[0])) )
        
    assert(sln.generators_status.ix[t0, 'g1'] == \
        value(generators[1].status(times[0])) )



'''Test for memory leaks'''
from minpower import powersystems, solve
from test_utils import solve_problem, make_loads_times
from coopr import pyomo


import objgraph, inspect

def show_memory_backrefs(name):    
    objgraph.show_backrefs(
        objgraph.by_type(name),
        filename='backrefs-{}.png'.format(name))


def show_memory_refs(name):
    try: obj=objgraph.by_type(name)[0]
    except IndexError:
        print 'no object of type',name  
        return
    objgraph.show_chain(
        objgraph.find_backref_chain(obj, inspect.ismodule),
        filename='chain-{}.png'.format(name))


def show_memory_growth():
    objgraph.show_growth()

def get_counts(prefix=''):
    test_types = [
        'Var', 
        'Piecewise',
        'ScenarioTree',
        'ScenarioTreeNode',
        'Scenario',
        '_SetContainer',
        '_ConstraintArray',
        ]
    
    # use objgraph to check if these pyomo objects still exist 
    objects = {} 
    test_counts = {}
    for name in test_types:
        objects[name] = objgraph.by_type( name )
        test_counts[name] = len(objects[name])
    
    if True:
        for name in test_types:
            if test_counts[name]==0: continue
            else: obj = objects[name][0]
            fname = prefix+'-'+'objgraph-{}-'.format(name)
            objgraph.show_refs( [obj], filename=fname+'refs.png') # too_many=50,
            objgraph.show_backrefs([obj], filename=fname+'backref.png')
            objgraph.show_chain(
                objgraph.find_backref_chain( obj, inspect.ismodule),
                filename=fname+'chain.png'
                )
    return test_counts    

def leak_on_reset():
    # create a problem
    loads_times = make_loads_times(Pdt=[20,30,40,50])
    generators = [
        powersystems.Generator(name='expensive', costcurveequation='30P+0.01P^2'),
        powersystems.Generator(name='cheap', costcurveequation='20P+0.005P^2')
        ]
    
    # solve it
    power_system, times = solve_problem(generators, **loads_times)
    # reset the model - no pyomo components should persist
    power_system.reset_model()
        
    test_counts = get_counts('uc')    
    assert sum(test_counts.values()) == 0

def leak_on_stochastic_reset():

    solve.solve_problem(datadir='./uc-stochastic',
        shell = False,
        csv = False,
        hours_commitment = 24,
        hours_commitment_overlap = 12,
        get_duals = False,
        Nscenarios = None,
    )   
    
    test_counts = get_counts('uc-stochastic')    
    assert sum(test_counts.values()) == 0


if __name__ == "__main__": 
    leak_on_reset()



'''Test the constraint behavior of the bids'''
from minpower.generators import Generator
from minpower.optimization import value

from test_utils import *


@istest
def linear():
    '''
    Create a linear bid curve for one generators.
    Ensure that correct cost is valued for the load.
    '''
    a=5
    b=30
    Pd=221
    generators=[ Generator(costcurveequation='{}+{}P'.format(a,b)) ]
    _,times=solve_problem(generators,**make_loads_times(Pd))
    cost = generators[0].bids.output(times[0], evaluate=True)
    assert cost == a + b*Pd

@istest
@with_setup(teardown=reset_config)
def cubic_convex():
    '''
    Create a cubic, convex bid curve for one generators.
    Ensure that linearized cost is within +5% of the true cost
    '''
    Pd=221
    a=5
    b=30
    c=.2
    d=.1
    user_config.breakpoints = 10
    
    generators=[ Generator(costcurveequation='{}+{}P+{}P^2+{}P^3'.format(a,b,c,d)) ]
    _,times=solve_problem(generators,**make_loads_times(Pd))#,problem_filename='bidproblem.lp')
    cost = value(generators[0].bids.output(times[0], evaluate=True))
    actual_cost = a+ b*Pd+ c*Pd**2 + d*Pd**3
    assert actual_cost <= cost and cost <= 1.05*actual_cost


@istest
def cubic_non_convex():
    '''
    Create a cubic, but non-convex (negative cubic term) bid curve for one generators.
    Ensure that linearized cost is within +5% of the true cost
    '''
    Pd=221
    a=5
    b=30
    c=.2
    d=.0001
    generators=[ Generator(costcurveequation='{}+{}P+{}P^2 - {}P^3'.format(a,b,c,d)) ]
    power_system,times=solve_problem(generators,**make_loads_times(Pd))

    cost = generators[0].bids.output(times[0],evaluate=True) 
    actual_cost = a+ b*Pd+ c*Pd**2 + -1*d*Pd**3
    assert actual_cost <= cost <= 1.05*actual_cost

@istest
def coverage():
    '''
    Make sure that the bid range covers the whole pmin-pmax range
    '''
    Pd=221
    a=5
    b=30
    c=.2
    d=.1
    pmin = 10
    pmax = 559
    generators=[ Generator(
        costcurveequation='{}+{}P+{}P^2+{}P^3'.format(a,b,c,d), 
        pmin=pmin, 
        pmax=pmax) ]
    _,times=solve_problem(generators,**make_loads_times(Pd))
    bid_points = generators[0].bids.discrete_input_points
    assert pmin == bid_points[0] and pmax == bid_points[-1]


@istest
def fixed_costs_when_off():
    '''
    ensure that generator with fixed cost 
    only charges fixed cost when on
    '''
    a = 5
    b = 30
    c = 0.2
    
    generators=[
        make_cheap_gen(pmax=80),
        make_mid_gen(pmax=20),
        make_expensive_gen(
            costcurveequation='{}+{}P+{}P^2'.format(a,b,c),
            mindowntime=1,
            pmax=50
            )
    ]

    Pdt = [80, 90, 130]

    power_system, times = solve_problem(
        generators, 
        gen_init=[
            {'power': 80}, 
            {'power': 0}, 
            {'status': False, 'hoursinstatus': 0}],
        **make_loads_times(Pdt=Pdt))
    
    assert(generators[2].cost(times[0], evaluate=True) == 0)
    
if __name__ == '__main__':
    nose.runmodule(argv=[__file__,'-vvs','-x','--pdb', '--pdb-failure'],
                   exit=False)



import numpy as np
from commonscripts import update_attributes, pairwise
from optimization import value, OptimizationObject
from config import user_config
import re
from coopr.pyomo import Piecewise


class Bid(OptimizationObject):
    """
    A bid modeled by a polynomial or a set of piecewise points.
    """
    def __init__(self,
                 polynomial='10P',
                 bid_points=None,
                 constant_term=0,
                 owner=None,
                 times=None,
                 input_variable=0,
                 min_input=0,
                 max_input=1000,
                 num_breakpoints=user_config.breakpoints,
                 status_variable=True,
                 fixed_input=False):
        update_attributes(self, locals(), exclude=['owner'])
        self._parent_problem = owner._parent_problem
        self.owner_id = str(owner)

        self.is_pwl = (self.bid_points is not None)
        self.is_linear = is_linear(self.polynomial)

        if not fixed_input:
            self.build_model()

    def build_model(self):
        if self.bid_points is None:

            self.is_pwl = False
            self.constant_term = self.polynomial[0]

            if self.is_linear:
                return

            polynomial = list(self.polynomial)
            # use constant term in place of the 0th order term
            polynomial[0] = 0

            self.add_variable('cost', index=self.times.set, low=0)

            def pw_rule(model, time, input_var):
                return polynomial_value(polynomial, input_var)
            self.discrete_input_points = discretize_range(
                self.num_breakpoints, self.min_input, self.max_input)
            in_pts = dict(
                (t, self.discrete_input_points) for t in self.times.set)

            pw_representation = Piecewise(
                self.times.set,
                self.get_variable('cost', time=None, indexed=True),
                self.input_variable(),
                f_rule=pw_rule,
                pw_pts=in_pts,
                pw_constr_type='LB',
                warn_domain_coverage=False,
                # unless warn_domain_coverage is set, pyomo will complain
                # gen lower power bounds are set to zero (status trick)
                # and Piecewise complains if Pmin>0,
            )

        else:
            # custom bid points
            self.is_linear = False
            self.is_pwl = True
            self.add_variable('cost', index=self.times.set, low=0)
            self.discrete_input_points = self.bid_points.power.values.tolist()
            in_pts = dict(
                (t, self.discrete_input_points) for t in self.times.set)
            mapping = self.bid_points.set_index('power').to_dict()['cost']

            def pw_rule_points(model, time, input_var):
                # just the input->output points mapping in this case
                # see coopr/examples/pyomo/piecewise/example3.py
                return mapping[input_var]

            pw_representation = Piecewise(self.times.set,
                                          self.get_variable(
                                          'cost', time=None, indexed=True),
                                          self.input_variable(),
                                          pw_pts=in_pts,
                                          pw_constr_type='LB',
                                          pw_repn='DCC',  # the disagregated convex combination method
                                          f_rule=pw_rule_points,
                                          warn_domain_coverage=False)

        pw_representation.name = self.iden()
        self.max_output = pw_representation._f_rule(None, None, self.max_input)
        self._parent_problem().add_component_to_problem(pw_representation)

    def output(self, time=None, scenario=None, evaluate=False):
        status = self.status_variable(time, scenario)
        power = self.input_variable(time, scenario)
        if evaluate:
            status = value(status)
            power = value(power)

        if self.is_linear:
            out = self.polynomial[1] * power
        else:
            out = self.get_variable('cost',
                                    time=time, scenario=scenario, indexed=True)
            if evaluate:
                out = value(out)

        if self.constant_term != 0:
            out += status * self.constant_term

        return out

    def output_true(self, input_var, force_linear=False):
        '''true output value of bid'''
        input_val = value(input_var)

        if (self.is_pwl or force_linear) and not self.is_linear:
            if not self.is_pwl and self.bid_points is None:
                # construct the bid points
                bid_pt_outputs = map(lambda pt: polynomial_value(
                    self.polynomial, pt), self.discrete_input_points)
                self.bid_points = zip(
                    self.discrete_input_points, bid_pt_outputs)

            for A, B in pairwise(self.bid_points.values.tolist()):
                if A[0] <= input_val <= B[0]:
                    return get_line_value(A, B, input_val) + self.constant_term
        else:
            return polynomial_value(self.polynomial, input_val)

    def output_incremental(self, input_var):
        input_val = value(input_var)
        if self.is_pwl:
            for A, B in pairwise(self.bid_points):
                if A[0] <= input_val <= B[0]:
                    return get_line_slope(A, B)
        else:
            return polynomial_incremental_value(self.polynomial, value(input_var))

    def output_incremental_range(self):
        if self.is_pwl:
            input_range = self.discrete_input_points
            output_range = [0]
            for A, B in pairwise(self.bid_points):
                output_range.append(get_line_slope(A, B))
        else:
            input_range = np.arange(self.min_input, self.max_input, 1.0)
            output_range = [polynomial_incremental_value(self.polynomial, x)
                            for x in input_range]
        return input_range, output_range

    def __str__(self):
        return 'bid_{}'.format(self.owner_id)

    def iden(self, *a, **k):
        return 'bid_{}'.format(self.owner_id)


def is_linear(coefs):
    result = False
    if coefs is None:
        result = False
    else:
        if len(coefs) < 2:
            result = True
        elif all(m == 0 for m in coefs[2:]): result = True
        else:
            result = False
    return result


def discretize_range(num_breakpoints, minimum, maximum):
    step = (maximum - minimum) / float(num_breakpoints - 1)
    return [x * step + minimum for x in range(int(num_breakpoints))]


def polynomial_value(multipliers, variable):
    """get the value of a polynomial"""
    def term(mult, var, order):
        if order > 1:
            return mult * variable ** order
        elif order == 1:
            return mult * variable
        elif order == 0:
            return mult
    return sum([term(mult, variable, order) for order, mult in enumerate(multipliers)])


def polynomial_incremental_value(multipliers, variable):
    """get the incremental value of a polynomial"""
    return sum([(mult * order * variable ** (order - 1) if order > 0 else 0) for order, mult in enumerate(multipliers)])


def parse_polynomial(s):
    """
    Parse a string into a set of multipliers.
    Heavily adapted from `<http://bit.ly/polynomialParse>`_.

    Can handle simple polynomials (addition and subtraction):

    >>> parse_polynomial('7x^2 + 6x - 5')
    [-5, 6, 7]

    or with the explicit * multiplier:

    >>> parse_polynomial('7*P^2 + 6*P - 5')
    [-5, 6, 7]

    or even with the terms in some random order:

    >>> parse_polynomial('6*P - 5 + 7*P^2')
    [-5, 6, 7]
    """

    def parse_n(s):
        '''Parse the number part of a polynomial string term'''
        if not s:
            return 1
        elif s == '-':
            return -1
        elif s == '+':
            return 1
        return float(eval(s))

    def parse_p(s, powerPattern):
        '''Parse the power part of a polynomial string term'''
        if not s:
            return 0
        multipliers = powerPattern.findall(s)[0]
        if not multipliers:
            return 1
        return int(multipliers)
    s = str(s).replace(' ', '')  # remove all whitespace from string
    m = re.search('[a-zA-Z]+', s)
    try:
        varLetter = m.group(0)
    except AttributeError:
        varLetter = 'P'
    termPattern = re.compile(
        '([+-]?\d*\.?\d*)\**({var}?\^?\d?)'.format(var=varLetter))
    powerPattern = re.compile('{var}\^?(\d)?'.format(var=varLetter))
    order_multipliers = {}

    for n, p in termPattern.findall(s):
        n, p = n.strip(), p.strip()
        if not n and not p:
            continue
        n, p = parse_n(n), parse_p(p, powerPattern)
        if p in order_multipliers:
            order_multipliers[p] += n
        else:
            order_multipliers[p] = n
    highest_order = max(
        max(order_multipliers.keys()), 1)  # order must be at least linear
    multipliers = [0] * (highest_order + 1)
    for key, value in order_multipliers.items():
        multipliers[key] = value

    return multipliers


def get_line_slope(A, B):
    xA, yA = A
    xB, yB = B
    return (yB - yA) * 1.0 / (xB - xA)


def get_line_value(A, B, x):
    '''
    take a pair of points and make a linear function
    get the value of the function at x
    see http://bit.ly/Pd4z4l
    '''
    xA, yA = A
    slope = get_line_slope(A, B)
    return slope * (value(x) - xA) + yA



import sys
import argparse
import subprocess
import os
from pprint import pprint
from minpower.config import (parse_command_line_config, 
    scheduler_config, user_config)

def main():
    '''
    a wrapper to make the minpower script work with cluster schedulers
    or with an ssh call that you start on your laptop and then close the
    connection.
    '''
    
    default_minpower_config = dict(user_config).copy()
    parser = argparse.ArgumentParser(
        description='Minpower scheduler command line interface')
    

    parser.add_argument('--scheduler_mode', 
        default=scheduler_config.scheduler_mode,
        choices=['qsub', 'nohup', 'pass'],
        help='''Mode of scheduler operation:
        qsub: use the qsub cluster scheduler
        nohup: use nohup, call as subprocess, and redirect stdin, stderr
        pass: just call minpower as a child process (for debugging)
        ''')
    parser.add_argument('--verbose', action="store_true",
        default=scheduler_config.verbose)
    parser.add_argument('--dry_run', action="store_true",
        help="just show, don't do", default=False)
    
    qsub_opts = parser.add_argument_group('qsub options',
        description='the following options only apply if using qsub')
    qsub_opts.add_argument('--email',
        default=scheduler_config.email,
        help='email on scheduler job completion')    
    qsub_opts.add_argument('--memory',
        type=int, default=scheduler_config.memory, 
        help='gigabytes of memory to limit job to')
    qsub_opts.add_argument('--hours_limit',
        type=int, default=scheduler_config.hours_limit, 
        help='hours of runtime to limit job to')        
    
    args, minpower_args_raw = parser.parse_known_args()
    
    minpower_parser = argparse.ArgumentParser('minpower')
    minpower_args = parse_command_line_config(minpower_parser,
        preparsed_args=minpower_args_raw)    
    
    # scheduler config can be loaded from the case directory
    parser.set_defaults(**dict(scheduler_config))
    args, __ = parser.parse_known_args()
    
    # subprocess style
    minpower_args['standalone'] = True
    minpower_args['pid'] = os.getpid()

    if args.scheduler_mode == 'qsub':
        # qsub makes all of its script calls from the home directory
        # so it requires an absolute path
        minpower_args['directory'] = os.path.abspath(
            minpower_args['directory'])
        
        if 'scenarios_directory' in minpower_args and minpower_args['scenarios_directory']:
            minpower_args['scenarios_directory'] = os.path.abspath(
                minpower_args['scenarios_directory'])
            	
    
    if args.verbose:
        pprint(minpower_args)

    scheduler_call = []
    minpower_call = ['minpower']
    stdout = sys.stdout
    stderr = subprocess.STDOUT

    def arg2str(k, v):
        if k == 'directory':
            s = v
        elif v == True:
            s = '--{k}'.format(k=k)
        else:
            s = '--{k}={v}'.format(k=k, v=v)
        return s
    # make a big chain of args
    minpower_call.extend(sorted(
        [arg2str(k, v) for k, v in minpower_args.iteritems() if
            (k == 'directory') or (
                (k in default_minpower_config) and \
                (v != default_minpower_config[k])
            )
        ]))
    
    mode = args.scheduler_mode
    if mode == 'pass':
        # just let all the commands pass through
        pass
    elif mode == 'nohup':
        scheduler_call = ['nohup']
        if args.dry_run:
            print("would have opened files to write" + \
                "\n\tstdout: {p}.out\n\tstderr: {p}.err".format(p=os.getpid()))
        else:
            stdout = open('{}.out'.format(os.getpid()),'w')
            stderr = open('{}.err'.format(os.getpid()),'w')
    elif mode == 'qsub':
        # see https://sig.washington.edu/itsigs/Hyak_Job_Scheduler
        # default walltime limit is only one hour - must set this
        scheduler_call = ['qsub',
            '-l nodes=1:ppn=12,feature=12core,mem={m}gb,walltime={h}:00:00'.format(
                m=args.memory, h=args.hours_limit),
            ]
                    
        if args.email:
            scheduler_call.extend([
                '-m ae', #mail on completion/failure
                '-M {e}'.format(e=args.email)
                ])

        # there should be a better way than having to call a script on disk
        script_name = './{}.sh'.format(os.getpid())
        if args.dry_run:
            print('would have written script {f}: \n{c}'.format(
                f=script_name,
                c=' '.join(minpower_call)))
        else:
            with open(script_name, 'w+') as f:
                f.write(' '.join(minpower_call))
        minpower_call = [script_name]

    # actually make the call
    if args.dry_run:
        print('would have executed as a {p}:\n{c}'.format(
            p='child process' if mode == 'pass' else 'subprocess',
            c=' '.join((scheduler_call + minpower_call))
            ))
    else:
        if mode == 'pass':
            subprocess.call(scheduler_call + minpower_call)
            pid = None            
        else:
            
            pid = subprocess.Popen(scheduler_call + minpower_call,
                stdout=stdout,
                stderr=stderr,
                ).pid

    if args.verbose:
        print 'parent process {}'.format(os.getpid())
        print 'starting run {}'.format(pid)


if __name__=='__main__': main()



"""
Power systems optimization problem solver
for ED, OPF, UC, and SCUC problems. The :mod:`solve`
module contains the top-level commands for creating
problems and solving them.
"""

import traceback
import sys, os, logging, subprocess
import time as timer
import argparse
import pdb

from config import user_config, parse_command_line_config
from commonscripts import joindir, StreamToLogger
import powersystems, get_data, stochastic, results
from standalone import store_times, init_store, get_storage, repack_storage

def _set_store_filename(pid=None):
    fnm = 'stage-store.hd5'
    if user_config.output_prefix or user_config.pid:
        fnm = '{}-{}'.format(pid, fnm)

    user_config.store_filename = joindir(user_config.directory, fnm)

def solve_multistage_standalone(power_system, times, scenario_tree, data):

    stage_times = times.subdivide(
        user_config.hours_commitment, user_config.hours_overlap)

    storage = init_store(power_system, stage_times, data)
    pid = user_config.pid if user_config.pid else os.getpid()
    has_pid = user_config.output_prefix or user_config.pid
    
    for stg,t_stage in enumerate(stage_times):
        logging.info('Stage starting at {}'.format(t_stage.Start.date()))
        # store current stage times
        storage = store_times(t_stage, storage)
        storage.close()
        storage = None
        command = 'standalone_minpower {dir} {stg} {pid} {db}'.format(
                dir=user_config.directory, stg=stg,
                pid='--pid {}'.format(pid) if has_pid else '',
                db='--debugger' if user_config.debugger else '')
        try: subprocess.check_call(command, shell=True, stdout=sys.stdout)
        except AttributeError:
            # HACK - avoid error when nose hijacks sys.stdout
            subprocess.check_call(command, shell=True)

    repack_storage()
    storage = get_storage()
    return storage, stage_times

def standaloneUC():
    '''the hook for the ``standalone_minpower`` script'''
    from standalone import store_state, load_state

    parser = argparse.ArgumentParser()
    parser.add_argument('directory', type=str, help='the problem direcory')
    parser.add_argument('stg', type=int, help='the stage number')
    parser.add_argument('--pid', default='',
        help='the process id of the parent')
    parser.add_argument('--debugger', action='store_true', default=False,
        help='do some debugging')        

    args = parser.parse_args()
    stg = args.stg
    user_config.directory = args.directory
    user_config.pid = args.pid

    try:          
        _set_store_filename(args.pid)    
        # load stage data
        power_system, times, scenario_tree = load_state()

        # override the stored config with the current command line config
        user_config.directory = args.directory
        user_config.debugger = args.debugger

        _setup_logging(args.pid)
      
        sln = create_solve_problem(power_system, times, scenario_tree, stage_number=stg)

        store = store_state(power_system, times, sln)
        store.close()
    except:
        if user_config.debugger or args.debugger:
            __, __, tb = sys.exc_info()
            traceback.print_exc()
            pdb.post_mortem(tb)            
        else:
            raise

    return

def solve_problem(datadir='.',
        shell=True,
        problemfile=False,
        csv=True,
        ):
    """
    Solve a optimization problem specified by spreadsheets in a directory.
    Read in data, create and solve the problem, and return solution.
    The problem type is determined by the data.
    All options are set within `user_config`.
    """
    user_config.directory = datadir
    
    pid = user_config.pid if user_config.pid else os.getpid()
        
    _set_store_filename(pid)
    _setup_logging(pid)


    logging.debug(dict(user_config))
    
    start_time = timer.time()
    logging.debug('Minpower reading {}'.format(datadir))
    generators, loads, lines, times, scenario_tree, data = get_data.parsedir()
    logging.debug('data read')
    power_system = powersystems.PowerSystem(generators,loads,lines)

    logging.debug('power system initialized')
    if times.spanhrs <= user_config.hours_commitment + user_config.hours_overlap:
        solution = create_solve_problem(power_system, times, scenario_tree)
    else: #split into multiple stages and solve
        if user_config.standalone:
            stage_solutions, stage_times = solve_multistage_standalone(
                power_system, times, scenario_tree, data)
        else:
            stage_solutions, stage_times = solve_multistage(
                power_system, times, scenario_tree, data)
        solution = results.make_multistage_solution(
            power_system, stage_times, stage_solutions)

    if shell:
        if user_config.output_prefix or user_config.pid:
            stdout = sys.stdout
            sys.stdout = StreamToLogger()
            solution.show()
            sys.stdout = stdout
        solution.show()
    if csv and not user_config.standalone: solution.saveCSV()
    if user_config.visualization: solution.visualization()
    logging.info('total time: {}s'.format(timer.time()-start_time))

    if user_config.on_complete_script:
        os.system(user_config.on_complete_script)

    return solution


def solve_multistage(power_system, times, scenario_tree=None, data=None):
    stage_times = times.subdivide(
        user_config.hours_commitment, user_config.hours_overlap)

    stage_solutions = []

    for stg, t_stage in enumerate(stage_times):
        logging.info('Stage starting at {}'.format(t_stage.Start.date()))
        # solve
        solution = create_solve_problem(
            power_system, t_stage, scenario_tree, stg)
        # add to stage solutions
        stage_solutions.append(solution)
        # reset model
        power_system.reset_model()
        # set inital state for next stage
        if stg < len(stage_times)-1:
            power_system.set_initialconditions(stage_times[stg+1].initialTime)

    return stage_solutions, stage_times


def create_solve_problem(power_system, times, scenario_tree=None,
    stage_number=None, rerun=False):
    '''create and solve an optimization problem.'''
    
    create_problem(power_system, times, scenario_tree,
        stage_number, rerun)
    
    instance = power_system.solve_problem(times)

    logging.debug('solved... get results')

    sln = results.make_solution(power_system, times)
    
    power_system.disallow_shedding()

    # resolve with observed power and fixed status
    if sln.is_stochastic:
        power_system.resolve_stochastic_with_observed(instance, sln)
    elif user_config.deterministic_solve or user_config.perfect_solve:
        power_system.resolve_determinisitc_with_observed(sln)
   
    if len(times)>1:
        power_system.get_finalconditions(sln)

        power_system.disallow_shedding()
        sln.stage_number = stage_number
    
    return sln

def create_problem(power_system, times, scenario_tree=None,
    stage_number=None, rerun=False):
    """Create an optimization problem."""

    logging.debug('initialized problem')
    power_system.create_variables(times)
    logging.debug('created variables')
    power_system.create_objective(times)
    logging.debug('created objective')
    power_system.create_constraints(times)
    logging.debug('created constraints')

    if scenario_tree is not None and sum(scenario_tree.shape) > 0 and not rerun:
        stochastic.construct_simple_scenario_tree(
            power_system, times, time_stage=stage_number)
        stochastic.define_stage_variables(power_system, times)
        stochastic.create_problem_with_scenarios(power_system, times)
    return

def _setup_logging(pid=None):
    ''' set up the logging to report on the status'''
    kwds = dict(
        level=user_config.logging_level if not user_config.debugger else logging.debug,
        datefmt='%Y-%m-%d %H:%M:%S',
        format='%(asctime)s %(levelname)s: %(message)s')
    # log to file if pid is set, unless in debugging mode
    if (user_config.output_prefix or user_config.pid) \
        and not user_config.debugger:
        kwds['filename'] = joindir(user_config.directory,
            '{}.log'.format(pid))
    if (user_config.logging_level > 10) and (not 'filename' in kwds):
        # don't log the time if debugging isn't turned on
        kwds['format'] = '%(levelname)s: %(message)s'
    logging.basicConfig(**kwds)


def main():
    '''
    The command line interface for minpower. For more info use:
    ``minpower --help``
    '''

    args = parse_command_line_config(
        argparse.ArgumentParser(description='Minpower command line interface'))
    
    directory = user_config.directory

    if not os.path.isdir(directory):
        msg = 'There is no folder named "{}".'.format(directory)
        raise OSError(msg)
    
    if args['profile']:
        print 'run profile'
        import cProfile
        prof = cProfile.Profile()
        prof.runcall(solve_problem, directory)
        prof.dump_stats('minpower.profile')

    else:
        #solve the problem with those arguments
        try: solve_problem(directory)
        except:
            if args['debugger']:
                __, __, tb = sys.exc_info()
                traceback.print_exc()
                pdb.post_mortem(tb)
            else: raise

# for use in dev
if __name__=='__main__': main()



"""
Get data from spreadsheet files and parse it into
:class:`~Generator`, :class:`~powersystems.Load`,
:class:`~powersystems.Bus`, and :class:`~powersystems.Line` objects.
Also extract the time information and create all
    :class:`~schedule.Timelist` objects.
"""
import pandas as pd
from pandas import DataFrame, Timestamp, read_csv
from glob import glob
from collections import OrderedDict
    
import powersystems
from schedule import (just_one_time, get_schedule,
                      TimeIndex, make_constant_schedule)
from commonscripts import (joindir, drop_case_spaces, set_trace)

from powersystems import PowerSystem
from generators import (Generator,
                        Generator_Stochastic, Generator_nonControllable)
from config import user_config

import os
import logging

fields = dict(
    Line=[
        'name',
        'tobus',
        'frombus',
        'reactance',
        'pmax'
    ],

    Generator=[
        'name',
        'kind',
        'bus',
        'pmin', 'pmax',
        'power',  # for a non-controllable gen in an ED
        'rampratemin', 'rampratemax',
        'minuptime', 'mindowntime',
        'startupramplimit', 'shutdownramplimit',
        # for a bid points defined gen, noloadcost replaces the constant
        # polynomial
        'noloadcost',
        'costcurveequation', 'heatrateequation',
        'fuelcost',
        'startupcost', 'shutdowncost',
        'faststart',
        'mustrun',
        'sheddingallowed',
    ],

    Load=['name', 'bus', 'power'],
)

# extra fields for generator scheduling and bid specification
# these fields require parsing of additional files
# and result in additional objects being added
# to the Generator after creation
gen_extra_fields = ['observedname', 'forecastname', 'scenariosfilename',
                    'scenariosdirectory', 'costcurvepointsfilename']

fields_initial = [
    'status',
    'power',
    'hoursinstatus']


def nice_names(df):
    '''drop the case and spaces from all column names'''
    return df.rename(columns=
                     dict([(col, drop_case_spaces(col)) for col in df.columns]))


def parse_standalone(storage, times):
    '''load problem info from a pandas.HDFStore'''

    # filter timeseries data to only this stage
    timeseries = storage['data_timeseries'].ix[times.strings.values]

    # add loads
    loads = build_class_list(storage['data_loads'], powersystems.Load,
                             times, timeseries)
    # add generators
    generators = build_class_list(storage['data_generators'], Generator,
                                  times, timeseries)
    # add lines
    lines = build_class_list(storage['data_lines'], powersystems.Line,
                             times, timeseries)

    power_system = PowerSystem(generators, loads, lines)

    gen = power_system.get_generator_with_scenarios()
    if gen:
        # TODO - maybe only extract current day's values here
        scenario_values = storage['data_scenario_values']
        gen.scenario_values = scenario_values
    else:
        scenario_values = pd.Panel()

    return power_system, times, scenario_values


def _load_raw_data():
    """import data from spreadsheets"""
    datadir = user_config.directory

    if not os.path.isdir(datadir):
        raise OSError('data directory "{d}" does not exist'.format(d=datadir))
    [file_gens, file_loads, file_lines, file_init] = \
        [joindir(datadir, filename) for filename in (
            user_config.file_gens,
            user_config.file_loads, 
            user_config.file_lines, 
            user_config.file_init)]

    generators_data = nice_names(read_csv(file_gens))
    loads_data = nice_names(read_csv(file_loads))

    try:
        lines_data = nice_names(read_csv(file_lines))
    except Exception:
        lines_data = DataFrame()

    try:
        init_data = nice_names(read_csv(file_init))
    except Exception:
        init_data = DataFrame()

    return generators_data, loads_data, lines_data, init_data


def _parse_raw_data(generators_data, loads_data, lines_data, init_data):
    # create times
    timeseries, times, generators_data, loads_data = setup_times(
        generators_data, loads_data)

    # add loads
    loads = build_class_list(loads_data, powersystems.Load, times, timeseries)

    # data modifiers
    if user_config.pmin_multiplier != 1.0:
        generators_data['pmin'] *= user_config.pmin_multiplier
        invalid_pmin = (generators_data.pmin > generators_data.pmax)
        if invalid_pmin.any():
            logging.warning('Pmin scaling resulted in Pmin>Pmax. ' +
                            'Setting Pmin=Pmax for these {} generators.'.format(
                                invalid_pmin.sum()))
            generators_data.ix[invalid_pmin, 'pmin'] = \
                generators_data.ix[invalid_pmin, 'pmax']

    if user_config.ramp_limit_multiplier != 1.0:
        generators_data['rampratemin'] *= user_config.ramp_limit_multiplier
        generators_data['rampratemax'] *= user_config.ramp_limit_multiplier

    if user_config.ignore_minhours_constraints:
        generators_data['minuptime'] = 0
        generators_data['mindowntime'] = 0
    if user_config.ignore_ramping_constraints:
        generators_data['rampratemax'] = None
        generators_data['rampratemin'] = None
    if user_config.ignore_pmin_constraints:
        generators_data['pmin'] = 0

    # add generators
    generators = build_class_list(
        generators_data, Generator, times, timeseries)

    # add lines
    lines = build_class_list(lines_data, powersystems.Line)
    # add initial conditions
    setup_initialcond(init_data, generators, times)

    # get scenario values (if applicable)
    scenario_values = setup_scenarios(generators_data, generators, times)

    # also return the raw DataFrame objects
    data = dict(
        generators=generators_data,
        loads=loads_data,
        lines=lines_data,
        init=init_data,
        timeseries=timeseries,
        scenario_values=scenario_values,
    )

    return generators, loads, lines, times, scenario_values, data


def parsedir(**filename_kwargs):
    """
    Import data from spreadsheets and build lists of
    :mod:`powersystems` classes.
    """

    generators_data, loads_data, lines_data, init_data = _load_raw_data(
        **filename_kwargs)

    return _parse_raw_data(generators_data, loads_data,
                           lines_data, init_data)


def setup_initialcond(data, generators, times):
    '''
    Take a list of initial conditions parameters and
    add information to each :class:`~Generator` object.
    '''
    if len(times) <= 1:
        return  # for UC,ED no need to set initial status

    t_init = times.initialTime
    if len(data) == 0:
        logging.warning('''No generation initial conditions file found.
            Setting to defaults.''')
        for gen in generators:
            gen.set_initial_condition(time=t_init)
        return

    # begin by setting initial condition for all generators to off
    for g in generators:
        g.set_initial_condition(t_init, status=False, power=0)

    names = [g.name for g in generators]

    if not 'name' in data.columns:
        # assume they are in order
        data['name'] = names

    if 'power' not in data.columns:
        raise KeyError('initial conditions file should contain "power".')

    # add initial conditions for generators
    # which are specified in the initial file
    for i, row in data.iterrows():
        g = names.index(row['name'])
        kwds = row[fields_initial].dropna().to_dict()
        generators[g].set_initial_condition(time=t_init, **kwds)
    return


def build_class_list(data, model, times=None, timeseries=None):
    """
    Create list of class instances from the row of a DataFrame.
    """
    datadir = user_config.directory
    is_generator = (model == Generator)

    all_models = []

    if 'schedulename' in data.columns:
        data['schedule'] = None

    for i, row in data.iterrows():
        row_model = model
        row = row.dropna()

        power = row.get('power')
        schedulename = row.get('schedulename')

        if is_generator:
            # get all those extra things which need to be parsed
            observed_name = row.get('observedname')
            forecast_name = row.get('forecastname')

            scenariosfilename = row.get('scenariosfilename')
            scenariosdirectory = row.get('scenariosdirectory')

            if scenariosdirectory and user_config.scenarios_directory:
                # override the scenarios directory with the one \
                # specified in the commandline options
                scenariosdirectory = user_config.scenarios_directory
                data.ix[i,
                        'scenariosdirectory'] = user_config.scenarios_directory

            bid_points_filename = row.get('costcurvepointsfilename')

        if is_generator:
            if schedulename or power or \
                (forecast_name and user_config.deterministic_solve) or \
                    (observed_name and user_config.perfect_solve):
                row_model = Generator_nonControllable
            elif scenariosdirectory or scenariosfilename:
                row_model = Generator_Stochastic

        # warn about fields not in model
        valid_fields = pd.Index(fields[model.__name__] + ['schedulename'])
        if is_generator:
            valid_fields = valid_fields.union(pd.Index(gen_extra_fields))
        invalid_fields = row.index.diff(valid_fields)
        if len(invalid_fields) > 0:
            raise ValueError('invalid fields in model:: {}'.format(
                invalid_fields.tolist()))
            # logging.warning

        kwds = row[row.index.isin(fields[model.__name__])].to_dict()

        # add in any schedules
        if schedulename:
            kwds['schedule'] = timeseries[schedulename]
        elif pd.notnull(power):
            # a constant power schedule
            kwds['schedule'] = make_constant_schedule(times, power)
            kwds.pop('power')

        if is_generator:
            if observed_name:
                kwds['observed_values'] = timeseries[observed_name]

            if user_config.perfect_solve and observed_name:
                # for a perfect information solve forecast = observed
                kwds['schedule'] = kwds['observed_values'] \
                    = timeseries[observed_name]
            elif forecast_name:
                kwds['schedule'] = timeseries[forecast_name]

            if scenariosdirectory:
                try:
                    kwds['observed_values'] = timeseries[observed_name]
                except:
                    raise IOError('''you must provide an
                        observed filename for a rolling stochastic UC''')

            # add a custom bid points file with {power, cost} columns
            if bid_points_filename:
                kwds['bid_points'] = read_bid_points(
                    joindir(datadir, bid_points_filename))
                kwds['costcurveequation'] = None

        try:
            obj = row_model(index=i, **kwds)
        except TypeError:
            print '{} model got unexpected parameter'.format(model)
            raise

        all_models.append(obj)

    return all_models


def read_bid_points(filename):
    bid_points = read_csv(filename)
    # return a dataframe of bidpoints
    return bid_points[['power', 'cost']].astype(float)


def setup_times(generators_data, loads_data):
    """
    Create a :class:`~schedule.TimeIndex` object
    from the schedule files.

    Also create a unified DataFrame of all the schedules, `timeseries`.

    If there are no schedule files (as in ED,OPF),
    create an index with just a single time.
    """
    fcol = 'schedulefilename'
    ncol = 'schedulename'

    loads_data[ncol] = None
    generators_data[ncol] = None

    if fcol not in loads_data.columns:
        loads_data[fcol] = None
    if fcol not in generators_data.columns:
        generators_data[fcol] = None

    datadir = user_config.directory

    timeseries = {}

    def filter_notnull(df, col):
        return df[df[col].notnull()]

    for i, load in filter_notnull(loads_data, fcol).iterrows():
        name = 'd{}'.format(i)
        loads_data.ix[i, ncol] = name
        timeseries[name] = get_schedule(joindir(datadir, load[fcol])) * \
            user_config.load_multiplier + user_config.load_adder

    for i, gen in filter_notnull(generators_data, fcol).iterrows():
        name = 'g{}'.format(i)
        generators_data.ix[i, ncol] = name
        timeseries[name] = get_schedule(joindir(datadir, gen[fcol]))

    # handle observed and forecast power
    fobscol = 'observedfilename'
    obscol = 'observedname'
    ffcstcol = 'forecastfilename'
    fcstcol = 'forecastname'

    obs_name = None    
    if fobscol in generators_data:
        generators_data[obscol] = None
        for i, gen in filter_notnull(generators_data, fobscol).iterrows():
            obs_name = 'g{}_observations'.format(i)
            generators_data.ix[i, obscol] = obs_name
            timeseries[obs_name] = get_schedule(joindir(datadir, gen[fobscol]))
            if user_config.wind_multiplier != 1.0:
                timeseries[obs_name] *= user_config.wind_multiplier
                
        generators_data = generators_data.drop(fobscol, axis=1)
    
    fcst_name = None
    if ffcstcol in generators_data:
        generators_data[fcstcol] = None
        for i, gen in filter_notnull(generators_data, ffcstcol).iterrows():
            fcst_name = 'g{}_forecast'.format(i)
            generators_data.ix[i, fcstcol] = fcst_name
            timeseries[fcst_name] = get_schedule(joindir(datadir, gen[ffcstcol])) * \
                user_config.wind_multiplier + user_config.wind_forecast_adder

            if user_config.wind_error_multiplier != 1.0:
                logging.debug('scaling wind forecast error')
                obs_name = 'g{}_observations'.format(i)
                error = timeseries[fcst_name] - timeseries[obs_name]
                timeseries[fcst_name] = timeseries[obs_name] + \
                    error * user_config.wind_error_multiplier

            if (timeseries[fcst_name] < 0).any():
                print timeseries[fcst_name].describe()
                logging.warning('Wind forecast must always be at least zero.')
                timeseries[fcst_name][timeseries[fcst_name] < 0] = 0
                
        generators_data = generators_data.drop(ffcstcol, axis=1)

    generators_data = generators_data.drop(fcol, axis=1)
    loads_data = loads_data.drop(fcol, axis=1)

    if len(timeseries) == 0:
        # this is a ED or OPF problem - only one time
        return DataFrame(), just_one_time(), generators_data, loads_data

    timeseries = DataFrame(timeseries)
    times = TimeIndex(timeseries.index)
    timeseries.index = times.strings.values

    if user_config.wind_capacity_factor != 0:
        if len(filter_notnull(generators_data, obscol)) != 1:
            raise NotImplementedError(
                'wind capacity factor only works with one wind generator')
            
        max_load = timeseries[filter(lambda col: col.startswith('d'),
            timeseries.columns)].sum(axis=1).max()
        capf_current = timeseries[obs_name].max() / max_load
        
        wind_mult = user_config.wind_capacity_factor / capf_current
        user_config.wind_multiplier = wind_mult

        timeseries[obs_name] *= wind_mult
        if fcst_name:
            timeseries[fcst_name] *= wind_mult

    return timeseries, times, generators_data, loads_data


def _parse_scenario_day(filename):
    logging.debug('reading scenarios from %s', filename)
    data = read_csv(filename, parse_dates=True, index_col=0)

    # select subset of scenarios
    Nscenarios = user_config.scenarios
    if Nscenarios:
        data = data[data.index < Nscenarios]
        data['probability'] = data['probability'] / sum(data['probability'])

    # return the data with probability column first
    return data[data.columns.drop('probability').insert(0, 'probability')]


def setup_scenarios(gen_data, generators, times):

    col = 'scenariosdirectory'
    scenario_values = pd.Panel()
    if user_config.deterministic_solve or user_config.perfect_solve or \
            (not col in gen_data.columns):
        # a deterministic problem
        return scenario_values

    gen_params = gen_data[gen_data[col].notnull()]
    if len(gen_params) > 1:
        raise NotImplementedError('more than one generator with scenarios.')

    gen = generators[gen_params.index[0]]
    gen.has_scenarios = True

    # directory of scenario values where each file is one day
    scenarios_directory = gen_params['scenariosdirectory'].values[0]

    searchstr = "*.csv"

    filenames = sorted(glob(joindir(user_config.directory,
                                    joindir(scenarios_directory, searchstr))))
    if not filenames:
        raise IOError('no scenario files in "{}"'.format(scenarios_directory))

    alldata = OrderedDict()
    for i, f in enumerate(filenames):
        data = _parse_scenario_day(f)
        # label scenarios for day with the first timestamp
        day = Timestamp(data.columns.drop('probability')[0])
        alldata[day] = data

    # TODO - assumes one hour intervals!!
    hrs = user_config.hours_commitment + user_config.hours_overlap

    # make scenarios into a pd.Panel with axes: day, scenario, {prob, [hours]}
    scenario_values = pd.Panel(
        items=alldata.keys(),
        major_axis=range(max([len(dat) for dat in alldata.values()])),
        minor_axis=['probability'] + range(hrs)
    )

    for day, scenarios in alldata.iteritems():
        if 'probability' == scenarios.columns[-1]:
            # reoder so that probability is the first column
            scenarios = scenarios[
                scenarios.columns[:-1].insert(0, 'probability')]
        # rename the times into just hour offsets
        scenarios = scenarios.rename(columns=dict(zip(scenarios.columns,
            ['probability'] + range(len(scenarios.columns) - 1))))

        # and take the number of hours needed
        scenarios = scenarios[scenarios.columns[:1 + hrs]]
        
        scenario_values[day] = scenarios

    if user_config.wind_multiplier != 1.0:
        scenario_values *= user_config.wind_multiplier
        svt = scenario_values.transpose(2, 1, 0)
        svt['probability'] *= 1 / user_config.wind_multiplier
        scenario_values = svt.transpose(2, 1, 0)

    gen.scenario_values = scenario_values
    # defer scenario tree construction until actual time stage starts
    return scenario_values


def _has_valid_attr(obj, name):
    return getattr(obj, name, None) is not None



from pandas import Series, DataFrame
from minpower.config import user_config
from minpower import powersystems, schedule, solve
from minpower.generators import Generator
from minpower.optimization import value, OptimizationError

import nose
from nose.tools import istest, with_setup, raises, set_trace
from nose.tools import eq_ as assert_equal


import logging
logging.basicConfig(
    level=logging.CRITICAL,
    format='%(levelname)s: %(message)s')


default_config = user_config.copy()


def get_duals():
    user_config.duals = True


def reset_config():
    # note - can't set a new user_config, because it would be a local variable
    #   instead, we update it with the defaults
    #   this only works if we aren't adding anything extra to the user_config
    user_config.update(default_config)
    try:
        assert(user_config == default_config)
    except:
        print DataFrame(dict(
            config=Series(user_config),
            default=Series(default_config)))
        raise

singletime = schedule.just_one_time()


gen_costs = dict(cheap=10, mid=20, expensive=30)


def assertAlmostEqual(x, y):
    assert(round(x - y, 5) == 0)


# def make_single_bus(generators,loads):
    # singlebus=powersystems.Bus()
    # singlebus.generators=generators
    # singlebus.loads=loads
    # return [singlebus]

def make_cheap_gen(**kwargs):
    return Generator(
        name='cheap gen',
        costcurveequation='{}P'.format(gen_costs['cheap']),
        **kwargs)


def make_mid_gen(**kwargs):
    return Generator(
        name='middle-range gen',
        costcurveequation='{}P'.format(gen_costs['mid']),
        **kwargs)


def make_expensive_gen(**kwargs):
    if 'costcurveequation' not in kwargs:
        kwargs['costcurveequation'] = '{}P'.format(gen_costs['expensive'])
    return Generator(name='expensive gen', **kwargs)


def make_loads_times(Pd=200, Pdt=None, **kwargs):
    if Pdt is None:
        times = singletime
        sched = Series(Pd, index=times)
        loads = [powersystems.Load(schedule=sched, **kwargs)]
    else:
        times = schedule.make_times_basic(N=len(Pdt))
        sched = Series(Pdt, index=times)
        loads = [powersystems.Load(schedule=sched, **kwargs)]

    return dict(loads=loads, times=times)


def solve_problem(generators, loads=None, times=None,
                  gen_init=None, lines=None, do_reset_config=True):

    if lines is None:
        lines = []

    if len(times) > 0:
        for g, gen in enumerate(generators):
            gen.index = g
            if gen_init is None:
                gen.set_initial_condition(times.initialTime)
            else:
                gen.set_initial_condition(times.initialTime, **gen_init[g])


    power_system = powersystems.PowerSystem(generators, loads, lines)
    solve.create_problem(power_system, times)

    try:
        instance = power_system.solve()
    except OptimizationError:
        # re-do stage, with load shedding allowed
        logging.critical('stage infeasible, re-running with load shedding.')
        power_system.allow_shedding(times)
        try:
            instance = power_system.solve()
        except OptimizationError:
            scheduled, committed = power_system.debug_infeasibe(times)
            power_system.write_model('infeasible.lp')
            raise OptimizationError(
                'failed to solve, even with load shedding.')

    if do_reset_config:
        reset_config()

    if power_system.solved:
        power_system.update_variables()
    else:
        power_system.write_model('problem.lp')
        raise OptimizationError('infeasible problem, wrote to problem.lp')
    return power_system, times



"""
Do an many economic dispatches over a range of load values to get an
'aggregate' cost curve for a system. 
"""


from minpower.config import user_config
from minpower.get_data import parsedir
from minpower.commonscripts import joindir
from minpower.tests.test_utils import make_loads_times, solve_problem
from pandas import DataFrame, Series, Index
import numpy as np
import matplotlib.pyplot as plt
import argparse

def main(args):
    generators, loads, _, times, _, data = parsedir()
    generators = filter(lambda gen: gen.is_controllable, generators)
    
    gen_data = data['generators']
    if args['min'] == 0:
        args['min'] = 1.1 * gen_data.pmin.sum()
        
    if args['max'] == 0:
        args['max'] = 0.99 * gen_data.pmax.sum()
    
    load_values = np.arange(args['min'], args['max'], args['interval'])
    results = DataFrame(columns=['prices', 'committed', 'last_committed'], index=load_values)
    
    committed_gen_names = Index([])
    
    for load_val in load_values:
        print load_val
        loads_times = make_loads_times(Pd=load_val)
        power_system, times = solve_problem(generators, 
            do_reset_config=False, **loads_times)
        t = times[0]
        results.ix[load_val, 'prices'] = power_system.buses[0].price(t)
        statuses = Series(dict([(gen.name, gen.status(t).value)
            for gen in power_system.generators()]))
        
        results.ix[load_val, 'committed'] = statuses.sum()
        results.ix[load_val, 'last_committed'] = \
            statuses[statuses == 1].index.diff(committed_gen_names)
        committed_gen_names = statuses[statuses == 1].index
    
    if (load_values[-1] == 0.99 * gen_data.pmax.sum()) and \
        (statuses.sum() != len(generators)):
        print('warning: uncommitted generation:')
        print(gen_data.set_index('name').ix[statuses[statuses==0].index])
    
    results.to_csv(joindir(user_config.directory, 'ed_sweep.csv'))
    
    if args['hide_units_committed']:
        ax = results.prices.plot(drawstyle='steps')
    else:
        ax = results[['prices', 'committed']].plot(drawstyle='steps', secondary_y=['committed'])
        ax.right_ax.set_ylabel('Units committed')

    ax.set_xlabel('System Load (MW)')
    ax.set_ylabel('Estimated System Price ($/MWh)')

    plt.savefig(joindir(user_config.directory, 'ed_sweep.png'))
    
def get_args():
    parser = argparse.ArgumentParser(
        description="""
        Do an many economic dispatches over a range of load values to get an
        'aggregate' cost curve for a system. """)
    parser.add_argument('directory', type=str,
        help='the direcory of the problem you want to solve')
    parser.add_argument('--min', type=int,
        default=0,
        help='the load value to start at')
    parser.add_argument('--max', type=int,
        default=0,
        help='the load value to end at')
    parser.add_argument('--interval', type=float,
        default=100,
        help='the interval to increment the load by')
    parser.add_argument('--hide_units_committed', '-c', 
        action='store_true', default=False)
    
    args = parser.parse_args()

    user_config.directory = args.directory
    user_config.duals = True
    user_config.perfect_solve = True  # hack - ignore scenarios
    user_config.dispatch_decommit_allowed = True
    
    return vars(args)

if __name__ == '__main__':
    args = get_args()
    main(args)



"""
An optimization command library for Minpower.
Basically a wrapper around Coopr's `pyomo.ConcreteModel` class.
"""
import logging
import time
import weakref
from commonscripts import quiet, not_quiet, set_trace

with quiet():    
    import coopr.pyomo as pyomo
from coopr.opt.base import solvers as cooprsolver
import pandas as pd

# make pyomo recognize that True == 1
pyomo.base.numvalue.KnownConstants[
    True] = pyomo.base.numvalue.NumericConstant(1.0)

variable_kinds = dict(
    Continuous=pyomo.Reals,
    Binary=pyomo.Boolean,
    Boolean=pyomo.Boolean)

from config import user_config
from commonscripts import update_attributes, joindir, set_trace


def full_filename(filename):
    return joindir(user_config.directory, filename)


class OptimizationObject(object):
    '''
    A base class for an optimization object.
    This also serves as a template for
    how :class:`~OptimizationObject`s are structured.
    '''
    def __init__(self, *args, **kwargs):
        '''
        Individual class defined.
        Initialize the object. Often this just means assigning
        all of the keyword arguments to self.
        The :meth:`~OptimizationObject.init_optimization` method
        should be called in __init__.
        '''
        update_attributes(self, locals())  # load in inputs
        self.init_optimization()

#    def get(self, attr):
#        '''
#        convinience method for getting attributes an object may not have
#        without throwing errors
#        '''
#        return getattr(self, attr, None)
    def init_optimization(self):
        '''
        Initialize optimization components: add a container for children.
        If the index is not defined, make it a hash of the object
        to ensure the index is unique.
        '''
        self.children = dict()
        if getattr(self, 'index', None) is None:
            self.index = hash(self)
        if getattr(self, 'name', None) == '':
            self.name = self.index + 1  # 1 and up naming

    def create_variables(self, times, *args, **kwargs):
        '''
        Individual class defined.
        Create the variables in this method by using calls to
        :meth:`~optimization.OptimiationObject.add_variable`.
        Variables will be accessible by using
        :meth:`~optimization.OptimiationObject.get_variable` (or by adding
        a shortcut methods, like :meth:`~powersystems.Generator.power`).
        '''
        return  # self.all_variables(times)

    def create_objective(self, times):
        '''
        Individual class defined.
        Return the contribution to the objective (cost) expression.
        :returns: an expression, the default is 0
        '''
        return 0

    def create_constraints(self, times, *args, **kwargs):
        '''
        Individual class defined.
        Create the constraints in this method by using calls to
        :meth:`~optimization.OptimiationObject.add_constraint`.
        Constraints will be accessible by using
        :meth:`~optimization.OptimiationObject.get_constraint` (or by adding
        a shortcut methods, like :meth:`~powersystems.Bus.price`.
        '''
        return  # self.all_constraints(times)


    def add_variable(self, name, time=None,
                     fixed_value=None,
                     index=None,
                     **kwargs):
        '''
        Create a new variable and add it to the object's variables and the model's variables.
        :param name: name of optimization variable.
        :param kind: type of variable, specified by string. {Continuous or Binary/Boolean}
        :param low: low limit of variable
        :param high: high limit of variable
        :param fixed_value: a fixed value for a variable (making it a parameter)
        :param time: a single time for a variable
        :param index: a :class:`pyomo.Set` over which a variable is created
        '''
        def map_args(kind='Continuous', low=None, high=None):
            return dict(bounds=(low, high), domain=variable_kinds[kind])
        orig_name = name
        if index is None:
            name = self._t_id(name, time)
            if fixed_value is None:
                var = pyomo.Var(name=name, **map_args(**kwargs))
                self._parent_problem().add_component_to_problem(var)
            else:
                var = pyomo.Param(name=name, default=fixed_value)
                # add var
                self._parent_problem().add_component_to_problem(var)
                # and set value
                var = self.get_variable(orig_name, time)
                var[None] = fixed_value
        else:
            name = self._id(name)

            if fixed_value is None:
                var = pyomo.Var(index, name=name, **map_args(**kwargs))
                self._parent_problem().add_component_to_problem(var)
            else:
                var = pyomo.Param(index, name=name, default=fixed_value)
                self._parent_problem().add_component_to_problem(var)
                var = self._parent_problem().get_component(name)
                for i in index:
                    var[i] = fixed_value

    def add_parameter(self, name, index=None, values=None, mutable=True, default=None, **kwargs):
        name = self._id(name)
        self._parent_problem().add_component_to_problem(
            pyomo.Param(index, name=name,
                        default=default, mutable=mutable, **kwargs))
        if values is not None:
            if pd.Series(values).count() != len(values):
                raise ValueError('a parameter value cannot be NaN')
            var = self._parent_problem().get_component(name)
            for i in index:
                var[i] = values[i]

    def add_constraint(self, name, time, expression):
        '''Create a new constraint and add it to the object's constraints and the model's constraints.'''
        cname = self._t_id(name, time)
        self._parent_problem().add_component_to_problem(
            pyomo.Constraint(name=cname, rule=expression))

    def add_constraint_set(self, name, index, expression):
        cname = self._id(name)
        self._parent_problem().add_component_to_problem(
            pyomo.Constraint(index, name=cname, rule=expression))
    
    def get_dual(self, cname, time=None):
        '''get the dual of a constraint of an LP problem''' 
        if user_config.duals:
            return self._parent_problem()._model.dual.getValue(
                self.get_constraint(cname, time))
        else:
            return None
    
    def get_variable(self, name, time=None, indexed=False, scenario=None):
        if indexed:
            var_name = self._id(name)
            if time is None:
                return self._parent_problem().get_component(var_name, scenario)
            else:
                index = str(time)
                return self._parent_problem().get_component(var_name, scenario)[index]
        else:
            var_name = self._t_id(name, time)
            return self._parent_problem().get_component(var_name, scenario)

    def get_constraint(self, name, time):
        return self._parent_problem().get_component(self._t_id(name, time))

    def get_parameter(self, name, time, indexed=False):
        if indexed:
            name = self._id(name)
            if time is None:
                return self._parent_problem().get_component(name)
            else:
                return self._parent_problem().get_component(name)[str(time)]
        else:
            return self._parent_problem().get_component(self._t_id(name, time))

    def add_children(self, objects, name):
        '''Add a child :class:`~optimization.OptimizationObject` to this object.'''
        self.children[name] = objects
        try:
            # if objects is actually a dictionary
            for child in self.children[name].values():
                child._parent_problem = self._parent_problem
        except AttributeError:
            for child in self.children[name]:
                child._parent_problem = self._parent_problem

        setattr(self, name, objects)
    # def add_component(self,obj):
    #     obj._parent_problem=self._parent_problem
    #     self.children[obj.name]=obj

    def get_child(self, name, time=None):
        '''
        Get a child :class:`~optimization.OptimizationObject`
        dependent on time from this object.
        '''
        try:
            if time is None:
                return self.children[name]
            else:
                return self.children[name][time]
        except KeyError:
            print self.children.keys()
            raise
    # def get_cost(self,times): return
    # self.objective+sum([child.get_cost(times) for child in self.children])

    def iden(self, time):
        '''
        Individual class defined.
        Identifing string for the object, depending on time.
        Used to name variables and constraints for the object.
        '''
        return str(self) + '_' + str(time)

    def _t_id(self, name, time):
        return name.replace(' ', '_') + '_' + self.iden(time)

    def _id(self, name):
        return name.replace(' ', '_') + '_' + str(self)

    def __str__(self):
        '''
        Individual class defined.
        A string representation of the object (used when calling ``print``).
        You probably want to override this one with a more descriptive string.
        '''
        return 'opt_obj{ind}'.format(ind=self.index)

    def _remove_component(self, name, time=None):
        key = self._t_id(name, time)
        delattr(self._parent_problem()._model, key)

    def values(self, name, reindex=None):
        '''return the values of an indexed pyomo component as a Series'''
        var = self._parent_problem().get_component(self._id(name))
        out = pd.Series(dict([(k, value(v)) for k, v in var.iteritems()]))
        if reindex is not None:
            out.index = reindex
        return out

class OptimizationProblem(OptimizationObject):
    '''an optimization problem/model based on pyomo'''
    def __init__(self):
        self.init_optimization()

    def init_optimization(self):
        self._model = pyomo.ConcreteModel('power system problem')
        self.stochastic_formulation = False
        self.solved = False
        self.children = dict()
        self.variables = dict()
        self.constraints = dict()

    def add_children(self, objL, name):
        '''Add a child :class:`~optimization.OptimizationObject` to this object.'''
        self.children[name] = objL
        setattr(self, name, objL)
        for child in self.children[name]:
            child._parent_problem = weakref.ref(self)

    def add_component_to_problem(self, component):
        '''add a optimization component to the model'''
        if ':' in component.name:
            raise ValueError('no colons allowed in optimization object names')
        self._model.add_component(component.name, component)

    def add_objective(self, expression, sense=pyomo.minimize):
        '''add an objective to the problem'''
        self._model.objective = pyomo.Objective(
            name='objective', rule=expression, sense=sense)

    def add_set(self, name, items, ordered=False):
        '''add a :class:`pyomo.Set` to the problem'''
        self._model.add_component(name,
                                   pyomo.Set(initialize=items, name=name, ordered=ordered))

    def add_variable(self, name, **kwargs):
        '''create a new variable and add it to the root problem'''
        def map_args(kind='Continuous', low=None, high=None):
            return dict(bounds=(low, high), domain=variable_kinds[kind])
        var = pyomo.Var(name=name, **map_args(**kwargs))
        self._model.add_component(name, var)

    def add_constraint(self, name, expression, time=None):
        cname = self._t_id(name, time) if time is not None else name
        self._model.add_component(cname,
                                   pyomo.Constraint(name=name, rule=expression))

    def add_suffix(self, name):
        self._model.add_component(name, 
            pyomo.Suffix(direction=pyomo.Suffix.IMPORT))


    def get_component(self, name, scenario=None):
        '''Get an optimization component'''
        if scenario is None:
            try:
                return getattr(self._model, name)
            except (AttributeError, KeyError) as NotInModelError:
                print 'error getting ', name
                self.show_model()
                raise
        else:
            try:
                return getattr(self._scenario_instances[scenario], name)
            except AttributeError:
                print 'error getting ', name
                self._scenario_instances[scenario].pprint()
                raise

    def write_model(self, filename):
        try:
            self._model.write(filename)
        except:
            self._model.pprint(filename)

    def _remove_component(self, name, time=None):
        key = self._t_id(name, time) if time is not None else name
        delattr(self._model, key)

    
    def reset_objective(self):
        delattr(self._model, 'objective')
    
    def reset_model(self):
        instances = [self._model]
        if self.stochastic_formulation:
            instances.append(self._stochastic_instance)

        for instance in instances:
#        piecewise models leak memory
# keep until Coopr release integrates:
# https://software.sandia.gov/trac/coopr/changeset/5781
            for pw in instance.active_components(pyomo.Piecewise).values():
                pw._constraints_dict = None
                pw._vars_dict = None
                pw._sets_dict = None

            # another memory leak
            for key, var in instance.active_components(pyomo.Param).iteritems():
                var._index = None
            for key, var in instance.active_components(pyomo.Var).iteritems():
                var.reset()
                var._index = None
                var._data = None
                # var._varval = None

                # var = None
                # if instance==self._stochastic_instance: debug()
                delattr(instance, key)

#        debug()
        # for stage in self._scenario_tree._stages

        if True and self.stochastic_formulation:
            # destroy scenario tree
            for stage in self._scenario_tree._stages:
                stage._cost_variable = None
                for node in stage._tree_nodes:
                    node._children = None
                    node._parent = None
                    node.__dict__ = {}
                else:
                    stage._tree_nodes = None
            else:
                self._scenario_tree._stages = None

            for scenario in self._scenario_tree._scenarios:
                scenario._leaf_node = None
                scenario._node_list = None
            else:
                self._scenario_tree._scenarios = None

            self._scenario_tree._tree_node_map = None
            self._scenario_tree._tree_nodes = None
            self._scenario_tree = None

            self._stochastic_instance.components._component = {}
            self._stochastic_instance.components._declarations = {}

            self._stochastic_instance = None

        self.solved = False
        self._model = pyomo.ConcreteModel()

    def show_model(self):
        components = self._model.components._component
        items = [pyomo.Set, pyomo.Param, pyomo.Var,
                 pyomo.Objective, pyomo.Constraint]
        for item in items:
            if not item in components:
                continue
            keys = components[item].keys()
            keys.sort()
            print len(keys), item.__name__ + " Declarations"
            for key in keys:
                components[item][key].pprint()
            print ""

    def update_variables(self):
        '''Replace the variables with their numeric value.'''
        for name, var in self._model.active_components(pyomo.Var).items():
            try:
                setattr(self._model, name, value(var))
            except ValueError:
                # for boolean sets this sometimes doesn't work due to rounding
                if var.domain == pyomo.Boolean:
                    setattr(self._model, name, round(value(var)))
                else:
                    raise

    def solve(self):
        '''
        Send the optimization problem off to the solver.
        '''

        solver = user_config.solver
        get_duals = user_config.duals

        logging.info('Solving with {s}'.format(s=solver))

        # create instance
        if self.stochastic_formulation:
            instance = self._stochastic_instance
        else:
            instance = self._model.create()
            logging.debug('... model created')

        results, elapsed = self._solve_instance(instance, solver)

        if self.solved:
            self.solution_time = elapsed  # results.Solver[0]['Wallclock time']
            logging.info('Problem solved in {}s.'.format(self.solution_time))

        if user_config.problem_file:
            self.write_model(full_filename('problem.lp'))

        if not self.solved:
            if user_config.problem_file and self.stochastic_formulation:
                self._stochastic_instance.pprint(full_filename(
                    'unsolved-stochastic-instance.txt'))
            raise OptimizationError('problem not solved')

        instance.load(results, allow_consistent_values_for_fixed_vars=True)
        logging.debug('... solution loaded')

        if get_duals:
            # resolve with fixed variables
            logging.info('resolving fixed-integer LP for duals')
            _fix_binary_variables(instance)

            results, elapsed = self._solve_instance(
                instance, get_duals=get_duals)
            self.solution_time += elapsed

            instance.load(results)
            logging.debug('... LP problem solved')

        if self.stochastic_formulation:
            self._scenario_tree.snapshotSolutionFromInstances(
                self._scenario_instances)

        self.objective = get_objective(results,
                                       name='MASTER' if self.stochastic_formulation else 'objective')

        # self.constraints = instance.active_components(pyomo.Constraint)
        # self.variables =   instance.active_components(pyomo.Var)
        return instance

    def __str__(self):
        return 'system'

    def _solve_instance(self, instance,
        solver=user_config.solver, 
        get_duals=False,
        keepfiles=False,
        ):
        
        if user_config.keep_lp_files: 
            keepfiles = True
        
        suffixes = ['dual'] if get_duals else []

        if not hasattr(self, '_opt_solver'):
            kwds = {}
#            try:
#                if solver == 'gurobi':
#                    import gurobipy
#                    kwds['solver_io'] = 'python'
#                elif solver == 'cplex':
#                    import cplex
#                    kwds['solver_io'] = 'python'
#            except ImportError: pass

            self._opt_solver = cooprsolver.SolverFactory(solver, **kwds)
            self._opt_solver.options.mipgap = user_config.mipgap
            

            if self._opt_solver is None:
                msg = 'solver "{}" not found by coopr'.format(solver)
                raise OptimizationError(msg)

        if user_config.solver_time_limit: 
            self._opt_solver.options.timelimit = user_config.solver_time_limit


        # if we are debugging, show the solver output
        show_solver_output = user_config.logging_level <= 10

        start = time.time()
         
        quiet_fn = not_quiet if keepfiles or show_solver_output else quiet
        
        with quiet_fn():
            results = self._opt_solver.solve(instance, 
                suffixes=suffixes, 
                keepfiles=keepfiles, 
                tee=show_solver_output,
                )
        try:
            self._opt_solver._symbol_map = None  # this should mimic the memory leak bugfix at: software.sandia.gov/trac/coopr/changeset/5449
        except AttributeError:
            pass  # should remove after this fix becomes part of a release
        elapsed = (time.time() - start)
        self.solved = detect_status(results, self._opt_solver.name)

        if self.solved and not get_duals:
            try:
                self.mipgap = results.Solution[0]['Gap']
                logging.debug('solution gap={}'.format(self.mipgap))
            except AttributeError:
                self.mipgap = None

        return results, elapsed

    def fix_binary_variables(self, fix_offs=True):
        _fix_binary_variables(
            self._model,
            self.stochastic_formulation,
            fix_offs)

    def _unfix_variables(self):
        _unfix_variables(self._model)
    def _fix_variables(self, names):
        _fix_variables(names, self._model)

    def _remove_all_constraints(self):
        for key in self._model.active_components(pyomo.Constraint).keys():
            delattr(self._model, key)


def _fix_binary_variables(instance, is_stochastic=False, fix_offs=True):
    '''fix binary variables to their solved values to create an LP problem'''
    active_vars = instance.active_components(pyomo.Var)
    for var in active_vars.values():
        if isinstance(var.domain, pyomo.base.IntegerSet) or \
                isinstance(var.domain, pyomo.base.BooleanSet):
            if var.is_indexed():
                for key, ind_var in var.iteritems():
                    if fix_offs or ind_var.value >= 1 - 1e-5:
                        # not quite integer values can create strange
                        # and often infeasible resolve problems
                        ind_var.value = round(ind_var.value)
                        ind_var.fixed = True
            else:
                if fix_offs or var.value == 1:
                    var.fixed = True
    if is_stochastic:
        for scenario_block in filter(
            lambda blk: type(blk) != pyomo.Piecewise,
            instance.active_components(pyomo.Block).values()
        ):
            _fix_binary_variables(scenario_block)
    # need to preprocess after fixing
    instance.preprocess()


def _fix_variables(names, instance):
    active_vars = instance.active_components(pyomo.Var)
    for var in active_vars.values():
        if var.name in names:
            if var.is_indexed():
                for key, ind_var in var.iteritems():
                    ind_var.fixed = True
            else:
                var.fixed = True
    instance.preprocess()


def _unfix_variables(instance):
    active_vars = instance.active_components(pyomo.Var)
    for var in active_vars.values():
        if var.is_indexed():
            for key, ind_var in var.iteritems():
                ind_var.fixed = False
        else:
            var.fixed = False


def value(variable):
    '''
    Value of an optimization variable after the problem is solved.
    If passed a numeric value, will return the number.
    '''
    try:
        return variable.value
    except AttributeError:
        return variable  # just a number


def detect_status(results, solver):
    '''decide between a solver success or failure'''
    status_text = str(results.solver[0]['Termination condition'])

    if status_text == 'optimal':
        success = True
    elif status_text in ['infeasible', 'unbounded']:
        success = False
    elif status_text in ['unknown', 'maxTimeLimit']:
        # max time limit - the solver may have reached a viable solution
        # unknown - an edge case encountered in resolves
        success = len(results.Solution.Variable.keys()) > 0
    else:
        success = False

    if not success:
        logging.critical('problem not solved' +
                         ' - solver terminated with status: "{}"'.format(status_text))
    return success


def get_objective(results, name='objective'):
    value = 0
    try:
        value = results.Solution.objective[name].value
    except AttributeError:
        objs = results.Solution.objective.values()
        value = [obj['Value'] for obj in objs if 'Value' in obj][0]
    return value


class OptimizationError(Exception):
    '''Error that occurs within solving an optimization problem.'''
    def __init__(self, ivalue):
        if ivalue:
            self.value = ivalue
        else:
            self.value = "Optimization Error: there was a problem"
        Exception.__init__(self, self.value)

    def __str__(self):
        return self.value


class OptimizationResolveError(OptimizationError):
    '''Error that occurs when re-solving an optimization problem.'''
    pass


class NotInModelError(Exception):
    '''
    Error that occurs when trying to find
    a component in the optimization model.
    '''
    def __init__(self, ivalue):
        if ivalue:
            self.value = ivalue
        else:
            self.value = '''the component you are looking
                for wasn't found in the model'''
        Exception.__init__(self, self.value)

    def __str__(self):
        return self.value
